"""Tests for JsonDataProtocol."""
#   Copyright 2019 WolkAbout Technology s.r.o.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
import json
import sys
import unittest

sys.path.append("..")  # noqa

from wolk_gateway_module.json_data_protocol import JsonDataProtocol
from wolk_gateway_module.model.actuator_command import (
    ActuatorCommand,
    ActuatorCommandType,
)
from wolk_gateway_module.model.actuator_state import ActuatorState
from wolk_gateway_module.model.actuator_status import ActuatorStatus
from wolk_gateway_module.model.alarm import Alarm
from wolk_gateway_module.model.configuration_command import (
    ConfigurationCommand,
    ConfigurationCommandType,
)
from wolk_gateway_module.model.message import Message
from wolk_gateway_module.model.sensor_reading import SensorReading


class JsonDataProtocolTests(unittest.TestCase):
    """JsonDataProtocol Tests."""

    DEVICE_PATH_PREFIX = "d/"
    REFERENCE_PATH_PREFIX = "r/"
    CHANNEL_WILDCARD = "#"
    CHANNEL_DELIMITER = "/"
    SENSOR_READING = "d2p/sensor_reading/"
    ALARM = "d2p/events/"
    ACTUATOR_SET = "p2d/actuator_set/"
    ACTUATOR_GET = "p2d/actuator_get/"
    ACTUATOR_STATUS = "d2p/actuator_status/"
    CONFIGURATION_SET = "p2d/configuration_set/"
    CONFIGURATION_GET = "p2d/configuration_get/"
    CONFIGURATION_STATUS = "d2p/configuration_get/"

    def test_get_inbound_topics_for_device(self):
        """Test that returned list is correct."""
        json_data_protocol = JsonDataProtocol()
        device_key = "some_key"

        self.assertEqual(
            json_data_protocol.get_inbound_topics_for_device(device_key),
            [
                self.ACTUATOR_SET
                + self.DEVICE_PATH_PREFIX
                + device_key
                + self.CHANNEL_DELIMITER
                + self.REFERENCE_PATH_PREFIX
                + self.CHANNEL_WILDCARD,
                self.ACTUATOR_GET
                + self.DEVICE_PATH_PREFIX
                + device_key
                + self.CHANNEL_DELIMITER
                + self.REFERENCE_PATH_PREFIX
                + self.CHANNEL_WILDCARD,
                self.CONFIGURATION_SET + self.DEVICE_PATH_PREFIX + device_key,
                self.CONFIGURATION_GET + self.DEVICE_PATH_PREFIX + device_key,
            ],
        )

    def test_is_actuator_get_message(self):
        """Test that message is actuator_get message."""
        json_data_protocol = JsonDataProtocol()

        message = Message(self.ACTUATOR_GET)

        self.assertTrue(json_data_protocol.is_actuator_get_message(message))

    def test_is_actuator_set_message(self):
        """Test that message is actuator_set message."""
        json_data_protocol = JsonDataProtocol()

        message = Message(self.ACTUATOR_SET)

        self.assertTrue(json_data_protocol.is_actuator_set_message(message))

    def test_is_configuration_set_message(self):
        """Test that message is configuration_set message."""
        json_data_protocol = JsonDataProtocol()

        message = Message(self.CONFIGURATION_SET)

        self.assertTrue(
            json_data_protocol.is_configuration_set_message(message)
        )

    def test_is_configuration_get_message(self):
        """Test that message is configuration_get message."""
        json_data_protocol = JsonDataProtocol()

        message = Message(self.CONFIGURATION_GET)

        self.assertTrue(
            json_data_protocol.is_configuration_get_message(message)
        )

    def test_make_actuator_set_command(self):
        """Test deserializing of actuator_set message."""
        json_data_protocol = JsonDataProtocol()

        message = Message(
            self.ACTUATOR_SET
            + self.DEVICE_PATH_PREFIX
            + "some_key"
            + self.CHANNEL_DELIMITER
            + self.REFERENCE_PATH_PREFIX
            + "REF",
            '{"value": "test"}',
        )

        expected = ActuatorCommand("REF", ActuatorCommandType.SET, "test")

        deserialized = json_data_protocol.make_actuator_command(message)

        self.assertEqual(expected, deserialized)

    def test_make_actuator_get_command(self):
        """Test deserializing of actuator_get message."""
        json_data_protocol = JsonDataProtocol()

        message = Message(
            self.ACTUATOR_GET
            + self.DEVICE_PATH_PREFIX
            + "some_key"
            + self.CHANNEL_DELIMITER
            + self.REFERENCE_PATH_PREFIX
            + "REF"
        )

        expected = ActuatorCommand("REF", ActuatorCommandType.GET)

        deserialized = json_data_protocol.make_actuator_command(message)

        self.assertEqual(expected, deserialized)

    def test_make_configuration_set_message(self):
        """Test deserializing of configuration_set message."""
        json_data_protocol = JsonDataProtocol()

        message = Message(
            self.CONFIGURATION_SET + self.DEVICE_PATH_PREFIX + "some_key",
            '{"config_1": "false"}',
        )

        expected = ConfigurationCommand(
            ConfigurationCommandType.SET, {"config_1": False}
        )

        deserialized = json_data_protocol.make_configuration_command(message)

        self.assertEqual(expected, deserialized)

    def test_make_configuration_large_data_size_set_message(self):
        """Test deserializing of configuration message with data size > 1."""
        json_data_protocol = JsonDataProtocol()

        message = Message(
            self.CONFIGURATION_SET + self.DEVICE_PATH_PREFIX + "some_key",
            '{"config_1": "2.3,3.4,4.4"}',
        )

        expected = ConfigurationCommand(
            ConfigurationCommandType.SET, {"config_1": (2.3, 3.4, 4.4)}
        )

        deserialized = json_data_protocol.make_configuration_command(message)

        self.assertEqual(expected, deserialized)

    def test_make_configuration_get_message(self):
        """Test deserializing of configuration_get message."""
        json_data_protocol = JsonDataProtocol()

        message = Message(
            self.CONFIGURATION_GET + self.DEVICE_PATH_PREFIX + "some_key"
        )

        expected = ConfigurationCommand(ConfigurationCommandType.GET)

        deserialized = json_data_protocol.make_configuration_command(message)

        self.assertEqual(expected, deserialized)

    def test_make_sensor_reading_message(self):
        """Test serializing of sensor reading for device key."""
        json_data_protocol = JsonDataProtocol()

        reading = SensorReading("REF", "value")

        expected = Message(
            self.SENSOR_READING
            + self.DEVICE_PATH_PREFIX
            + "some_key"
            + self.CHANNEL_DELIMITER
            + self.REFERENCE_PATH_PREFIX
            + "REF",
            '{"data": "value"}',
        )

        serialized = json_data_protocol.make_sensor_reading_message(
            "some_key", reading
        )

        self.assertEqual(expected, serialized)

    def test_make_multi_sensor_reading_message(self):
        """Test serializing of tuple sensor reading for device key."""
        json_data_protocol = JsonDataProtocol()

        reading = SensorReading("REF", (1, 2, 3))

        expected = Message(
            self.SENSOR_READING
            + self.DEVICE_PATH_PREFIX
            + "some_key"
            + self.CHANNEL_DELIMITER
            + self.REFERENCE_PATH_PREFIX
            + "REF",
            '{"data": "1,2,3"}',
        )

        serialized = json_data_protocol.make_sensor_reading_message(
            "some_key", reading
        )

        self.assertEqual(expected, serialized)

    def test_make_alarm_message(self):
        """Test serializing of alarm event for device key."""
        json_data_protocol = JsonDataProtocol()

        reading = Alarm("REF", True, 1557150524022)

        expected = Message(
            self.ALARM
            + self.DEVICE_PATH_PREFIX
            + "some_key"
            + self.CHANNEL_DELIMITER
            + self.REFERENCE_PATH_PREFIX
            + "REF",
            '{"data": "true", "utc": 1557150524022}',
        )

        serialized = json_data_protocol.make_alarm_message("some_key", reading)

        self.assertEqual(expected, serialized)

    def test_make_actuator_status_message(self):
        """Test serializing of actuator status for device key."""
        json_data_protocol = JsonDataProtocol()

        device_key = "some_key"
        reference = "REF"
        actuator_status = ActuatorStatus(reference, ActuatorState.READY, 15)

        expected = Message(
            self.ACTUATOR_STATUS
            + self.DEVICE_PATH_PREFIX
            + device_key
            + self.CHANNEL_DELIMITER
            + self.REFERENCE_PATH_PREFIX
            + reference,
            json.dumps({"status": "READY", "value": "15"}),
        )

        serialized = json_data_protocol.make_actuator_status_message(
            device_key, actuator_status
        )

        self.assertEqual(expected, serialized)

    def test_make_configuration_message(self):
        """Test serializing of configuration message for device key."""
        json_data_protocol = JsonDataProtocol()

        device_key = "some_key"
        configuration = {
            "ref1": False,
            "ref2": 2,
            "ref3": 4.4,
            "ref4": ("a", "b"),
        }

        expected = Message(
            self.CONFIGURATION_STATUS + self.DEVICE_PATH_PREFIX + device_key,
            '{"values": {"ref1": "false", "ref2": "2",'
            + ' "ref3": "4.4", "ref4": "a,b"}}',
        )

        serialized = json_data_protocol.make_configuration_message(
            device_key, configuration
        )

        self.assertEqual(expected, serialized)


if __name__ == "__main__":
    unittest.main()



"""Stub method for setting new actuator value."""
#   Copyright 2019 WolkAbout Technology s.r.o.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
from typing import Union


def handle_actuation(
    device_key: str, reference: str, value: Union[bool, int, float, str]
) -> None:
    """
    Set device actuator identified by reference to value.

    Must be implemented as non blocking.
    Must be implemented as thread safe.

    :param device_key: Device identifier
    :type device_key: str
    :param reference: Reference of the actuator
    :type reference: str
    :param value: Value to which to set the actuator
    :type value: str
    """
    pass



"""Reading type used for registering devices on WolkAbout IoT Platform."""
#   Copyright 2019 WolkAbout Technology s.r.o.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
from typing import Optional
from typing import Union

from wolk_gateway_module.model.data_type import DataType
from wolk_gateway_module.model.reading_type_measurement_unit import (
    ReadingTypeMeasurementUnit as Unit,
)
from wolk_gateway_module.model.reading_type_name import ReadingTypeName as Name


class ReadingType:
    """
    Reading type used for registering sensors on WolkAbout IoT Platform.

    Define a reading type for sensors,
    either a generic type by specifying
    a ``DataType`` (numeric, boolean or string) or entering
    a predefined one by using the enumerations provided in
    ``ReadingTypeName`` and ``ReadingTypeMeasurementUnit`` .

    Custom reading types can be used by passing string values
    for the name and measurement unit.

    :ivar name: Name of reading type
    :vartype name: Union[ReadingTypeName, str]
    :ivar unit: Measurement unit of reading type
    :vartype unit: Union[ReadingTypeMeasurementUnit, str]
    """

    def __init__(
        self,
        data_type: Optional[DataType] = None,
        name: Optional[Union[Name, str]] = None,
        unit: Optional[Union[Unit, str]] = None,
    ):
        """
        Reading type used for registering device's sensors.

        :param data_type: Data type for generic reading type
        :type data_type: Optional[DataType]
        :param name: Reading type name from defined enumeration or string for custom
        :type name: Optional[Union[ReadingTypeName, str]]
        :param unit: Reading type measurement unit from defined enumeration or string for custom
        :type unit: Optional[Union[ReadingTypeMeasurementUnit, str]]

        :raises ValueError: Unable to create a reading type from given input
        """
        if not data_type and not name and not unit:
            raise ValueError("Nothing passed, can't create reading type")

        if data_type:
            if not isinstance(data_type, DataType):
                raise ValueError("Invalid data type given")
            if data_type == DataType.NUMERIC:
                self.name: Union[Name, str] = Name.GENERIC
                self.unit: Union[Unit, str] = Unit.NUMERIC
            elif data_type == DataType.BOOLEAN:
                self.name = Name.GENERIC_BOOLEAN
                self.unit = Unit.BOOLEAN
            elif data_type == DataType.STRING:
                self.name = Name.GENERIC_TEXT
                self.unit = Unit.TEXT
            return

        if not (name and (isinstance(unit, str) or isinstance(unit, Unit))):
            raise ValueError("Both name and unit must be provided")

        if not ReadingType.validate(name, unit):
            raise ValueError("Invalid reading type name or unit")

        self.name = name
        self.unit = unit

    def __repr__(self) -> str:
        """
        Make string representation of reading type.

        :returns: representation
        :rtype: str
        """
        return f"ReadingType(name='{self.name}', unit='{self.unit}')"

    @staticmethod
    def validate(name: Union[Name, str], unit: Union[Unit, str]) -> bool:
        """
        Validate reading type name and measurement unit.

        :param name: Reading type name
        :type name: Union[ReadingTypeName, str]
        :param unit: Reading type measurement unit
        :type unit: Union[ReadingTypeMeasurementUnit, str]

        :returns: valid
        :rtype: bool
        """
        if not (isinstance(name, Name) or type(name) == str):
            return False
        if not (isinstance(unit, Unit) or type(unit) == str):
            return False

        if type(name) == str and type(unit) == str:
            return True

        if name == Name.GENERIC:
            return (
                True
                if unit == Unit.NUMERIC
                or unit == Unit.BIT
                or unit == Unit.PERCENT
                or unit == Unit.CO2_MOL
                or unit == Unit.X10C
                or unit == Unit.X100V
                or unit == Unit.X10PA
                else False
            )
        elif name == Name.TEMPERATURE:
            return (
                True
                if unit == Unit.KELVIN
                or unit == Unit.CELSIUS
                or unit == Unit.FAHRENHEIT
                or unit == Unit.CELSIUS_X2
                or unit == Unit.CELSIUS_X10
                else False
            )
        elif name == Name.PRESSURE:
            return (
                True
                if unit == Unit.PASCAL
                or unit == Unit.MILLIMETER_OF_MERCURY
                or unit == Unit.INCH_OF_MERCURY
                or unit == Unit.BAR
                or unit == Unit.ATMOSPHERE
                or unit == Unit.MILLIBAR
                or unit == Unit.PRESSURE_PERCENT
                or unit == Unit.MILLIBAR_X10
                or unit == Unit.MICROBAR
                or unit == Unit.KILO_PASCAL
                else False
            )
        elif name == Name.HUMIDITY:
            return (
                True
                if unit == Unit.HUMIDITY_PERCENT
                or unit == Unit.HUMIDITY_PERCENT_X10
                else False
            )
        elif name == Name.BATTERY_VOLTAGE:
            return (
                True
                if unit == Unit.VOLT
                or unit == Unit.MILLIVOLT
                or unit == Unit.CENTIVOLT
                else False
            )
        elif name == Name.MOVEMENT:
            return True if unit == Unit.MOVEMENT else False
        elif name == Name.LIGHT:
            return True if unit == Unit.LIGHT_PERCENT else False
        elif name == Name.ACCELEROMETER:
            return (
                True
                if unit == Unit.METRES_PER_SQUARE_SECOND
                or unit == Unit.GRAVITY
                else False
            )
        elif name == Name.GYROSCOPE:
            return True if unit == Unit.GYROSCOPE else False
        elif name == Name.LOCATION:
            return True if unit == Unit.LOCATION else False
        elif name == Name.HEART_RATE:
            return True if unit == Unit.BEATS_PER_MINUTE else False
        elif name == Name.BATTERY_POWER:
            return (
                True
                if unit == Unit.BATTERY or unit == Unit.BATTERY_X1000
                else False
            )
        elif name == Name.BREATHING_RATE:
            return True if unit == Unit.BREATHS_PER_MINUTE else False
        elif name == Name.CALORIES:
            return True if unit == Unit.CALORIES else False
        elif name == Name.ELECTRIC_CURRENT:
            return (
                True
                if unit == Unit.AMPERE or unit == Unit.MILLIAMPERE
                else False
            )
        elif name == Name.POWER:
            return (
                True
                if unit == Unit.WATT
                or unit == Unit.HORSEPOWER
                or unit == Unit.MILLIWATT
                else False
            )
        elif name == Name.FLOOR_POSITION:
            return True if unit == Unit.METER else False
        elif name == Name.FLUID_VOLUME:
            return (
                True
                if unit == Unit.OUNCE_LIQUID_UK
                or unit == Unit.OUNCE_LIQUID_US
                or unit == Unit.LITRE
                or unit == Unit.MILLILITRE
                or unit == Unit.GALLON_UK
                or unit == Unit.GALLON_DRY_US
                else False
            )
        elif name == Name.LENGHT:
            return (
                True
                if unit == Unit.METRE
                or unit == Unit.MILE
                or unit == Unit.FOOT
                or unit == Unit.POINT
                or unit == Unit.INCH
                or unit == Unit.PARSEC
                or unit == Unit.YARD
                or unit == Unit.MILLIMETER
                or unit == Unit.CENTIMETER
                or unit == Unit.KILOMETER
                else False
            )
        elif name == Name.MASS:
            return (
                True
                if unit == Unit.KILOGRAM
                or unit == Unit.GRAM
                or unit == Unit.MILLIGRAM
                or unit == Unit.METRIC_TON
                or unit == Unit.ATOMIC_MASS
                or unit == Unit.GALLON_LIQUID_US
                or unit == Unit.TON_UK
                or unit == Unit.TON_US
                or unit == Unit.POUND
                or unit == Unit.OUNCE
                or unit == Unit.ELECTRON_MASS
                else False
            )
        elif name == Name.SOUND_LEVEL:
            return True if unit == Unit.DECIBEL else False
        elif name == Name.SPEED:
            return (
                True
                if unit == Unit.KNOT
                or unit == Unit.KILOMETERS_PER_HOUR
                or unit == Unit.MILES_PER_HOUR
                or unit == Unit.MACH
                or unit == Unit.SPEED_OF_LIGHT
                or unit == Unit.METER_PER_SECOND
                else False
            )
        elif name == Name.TIME:
            return (
                True
                if unit == Unit.SECOND
                or unit == Unit.MINUTE
                or unit == Unit.HOUR
                or unit == Unit.MONTH
                or unit == Unit.DAY
                or unit == Unit.WEEK
                or unit == Unit.YEAR
                else False
            )
        elif name == Name.MAGNETIC_FLUX_DENSITY:
            return (
                True
                if unit == Unit.TESLA
                or unit == Unit.GAUSS
                or unit == Unit.MICRO_TESLA
                else False
            )
        elif name == Name.RADIATION:
            return (
                True
                if unit == Unit.SIEVERT
                or unit == Unit.BECQUEREL
                or unit == Unit.RUTHERFORD
                or unit == Unit.ROENTGEN
                or unit == Unit.RADIATION_DOSE_EFFECTIVE
                or unit == Unit.CURIE
                else False
            )
        elif name == Name.FORCE:
            return (
                True
                if unit == Unit.NEWTON
                or unit == Unit.POUND_FORCE
                or unit == Unit.GRAVITY_FORCE
                or unit == Unit.KILOGRAM_FORCE
                or unit == Unit.DYNE
                else False
            )
        elif name == Name.MEASURE:
            return (
                True
                if unit == Unit.SQUARE_METRE
                or unit == Unit.CUBIC_METRE
                or unit == Unit.BYTE
                or unit == Unit.GRADE
                or unit == Unit.HECTARE
                or unit == Unit.CUBIC_INCH
                or unit == Unit.REVOLUTION
                or unit == Unit.CENTIRADIAN
                or unit == Unit.RAD
                or unit == Unit.COMPUTER_POINT
                or unit == Unit.DEGREE_ANGLE
                or unit == Unit.SECOND_ANGLE
                or unit == Unit.MINUTE_ANGLE
                or unit == Unit.SPHERE
                or unit == Unit.ARE
                else False
            )
        elif name == Name.ANGLE:
            return True if unit == Unit.RADIAN else False

        elif name == Name.FREQUENCY:
            return (
                True
                if unit == Unit.HERTZ
                or unit == Unit.MEGA_HERTZ
                or unit == Unit.GIGA_HERTZ
                else False
            )
        elif name == Name.MAGNETIC_FLUX:
            return (
                True if unit == Unit.WEBER or unit == Unit.MAXWELL else False
            )
        elif name == Name.ELECTRIC_CAPACITY:
            return (
                True if unit == Unit.FARAD or unit == Unit.FARADAY else False
            )
        elif name == Name.ELECTRIC_RESISTANCE:
            return True if unit == Unit.OHM else False
        elif name == Name.ELECTRIC_MAGNETISM:
            return True if unit == Unit.GILBERT else False
        elif name == Name.ELECTRIC_ENERGY:
            return (
                True
                if unit == Unit.JOULE or unit == Unit.ELECTRON_VOLT
                else False
            )
        elif name == Name.ELECTRIC_INDUCTANCE:
            return True if unit == Unit.HENRY else False
        elif name == Name.ELECTRIC_CONDUCTANCE:
            return True if unit == Unit.SIEMENS else False
        elif name == Name.LUMINOUS_FLUX:
            return True if unit == Unit.LUMEN else False
        elif name == Name.LUMINOUS_INTENSITY:
            return True if unit == Unit.CANDELA else False
        elif name == Name.ILLUMINANCE:
            return True if unit == Unit.LUX else False
        elif name == Name.GENERIC_TEXT:
            return True if unit == Unit.TEXT else False
        elif name == Name.GENERIC_BOOLEAN:
            return True if unit == Unit.BOOLEAN else False

        assert False



"""Contains the Wolk class that ties together the whole package."""
#   Copyright 2019 WolkAbout Technology s.r.o.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
from inspect import signature
from reprlib import recursive_repr
from time import sleep
from typing import Callable
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union

from wolk_gateway_module.connectivity.connectivity_service import (
    ConnectivityService,
)
from wolk_gateway_module.interface.firmware_handler import FirmwareHandler
from wolk_gateway_module.json_data_protocol import JsonDataProtocol
from wolk_gateway_module.json_firmware_update_protocol import (
    JsonFirmwareUpdateProtocol,
)
from wolk_gateway_module.json_registration_protocol import (
    JsonRegistrationProtocol,
)
from wolk_gateway_module.json_status_protocol import JsonStatusProtocol
from wolk_gateway_module.logger_factory import logger_factory
from wolk_gateway_module.model.actuator_state import ActuatorState
from wolk_gateway_module.model.actuator_status import ActuatorStatus
from wolk_gateway_module.model.alarm import Alarm
from wolk_gateway_module.model.device import Device
from wolk_gateway_module.model.device_registration_request import (
    DeviceRegistrationRequest,
)
from wolk_gateway_module.model.device_status import DeviceStatus
from wolk_gateway_module.model.firmware_update_status import (
    FirmwareUpdateState,
)
from wolk_gateway_module.model.firmware_update_status import (
    FirmwareUpdateStatus,
)
from wolk_gateway_module.model.message import Message
from wolk_gateway_module.model.sensor_reading import SensorReading
from wolk_gateway_module.mqtt_connectivity_service import (
    MQTTConnectivityService,
)
from wolk_gateway_module.outbound_message_deque import OutboundMessageDeque
from wolk_gateway_module.persistence.outbound_message_queue import (
    OutboundMessageQueue,
)
from wolk_gateway_module.protocol.data_protocol import DataProtocol
from wolk_gateway_module.protocol.firmware_update_protocol import (
    FirmwareUpdateProtocol,
)
from wolk_gateway_module.protocol.registration_protocol import (
    RegistrationProtocol,
)
from wolk_gateway_module.protocol.status_protocol import StatusProtocol

Configuration = Dict[
    str,
    Union[
        int,
        float,
        bool,
        str,
        Tuple[int, int],
        Tuple[int, int, int],
        Tuple[float, float],
        Tuple[float, float, float],
        Tuple[str, str],
        Tuple[str, str, str],
    ],
]


class Wolk:
    """
    Core of this package, tying together all features.

    :ivar actuation_handler: Set new actuator values for your devices
    :vartype actuation_handler: Optional[Callable[[str, str,str], None]]
    :ivar actuator_status_provider: Get device's current actuator state
    :vartype actuator_status_provider: Optional[Callable[[str, str], Tuple[ActuatorState, Union[bool, int, float, str]]]]
    :ivar configuration_handler: Set new configuration values for your devices
    :vartype configuration_handler: Optional[Callable[[str, Dict[str, str]], None]]
    :ivar configuration_provider: Get device's current configuration options
    :vartype configuration_provider: Optional[Callable[[str],Dict[]]
    :ivar connectivity_service: Service that enables connection to WolkGateway
    :vartype connectivity_service: ConnectivityService
    :ivar data_protocol: Parse messages related to device data
    :vartype data_protocol: DataProtocol
    :ivar device_status_provider: Get device's current status
    :vartype device_status_provider: Callable[[str], DeviceStatus]
    :ivar devices: List of devices added to module
    :vartype devices: List[Device]
    :ivar firmware_handler: Handle commands related to firmware update
    :vartype firmware_handler: Optional[FirmwareHandler]
    :ivar firmware_update_protocol: Parse messages related to firmware update
    :vartype firmware_update_protocol: FirmwareUpdateProtocol
    :ivar host: WolkGateway's host address
    :vartype host: str
    :ivar log: Logger instance
    :vartype log: logging.Logger
    :ivar module_name: Name of module used for identification on WolkGateway
    :vartype module_name: str
    :ivar outbound_message_queue: Means of storing messages
    :vartype outbound_message_queue: OutboundMessageQueue
    :ivar port: WolkGateway's connectivity port
    :vartype port: int
    :ivar registration_protocol: Parse messages related to device registration
    :vartype registration_protocol: RegistrationProtocol
    :ivar status_protocol: Parse messages related to device status
    :vartype status_protocol: StatusProtocol
    """

    def __init__(
        self,
        host: str,
        port: int,
        module_name: str,
        device_status_provider: Callable[[str], DeviceStatus],
        actuation_handler: Optional[
            Callable[[str, str, Union[bool, int, float, str]], None]
        ] = None,
        actuator_status_provider: Optional[
            Callable[
                [str, str], Tuple[ActuatorState, Union[bool, int, float, str]]
            ]
        ] = None,
        configuration_handler: Optional[
            Callable[[str, Configuration], None]
        ] = None,
        configuration_provider: Optional[
            Callable[[str], Configuration]
        ] = None,
        firmware_handler: Optional[FirmwareHandler] = None,
        connectivity_service: Optional[ConnectivityService] = None,
        data_protocol: Optional[DataProtocol] = None,
        firmware_update_protocol: Optional[FirmwareUpdateProtocol] = None,
        registration_protocol: Optional[RegistrationProtocol] = None,
        status_protocol: Optional[StatusProtocol] = None,
        outbound_message_queue: Optional[OutboundMessageQueue] = None,
    ):
        """
        Construct an instance ready to communicate with WolkGateway.

        :param host: Host address of WolkGateway
        :type host: str
        :param port: TCP/IP port of WolkGateway
        :type port: int
        :param module_name: Module identifier used when connecting to gateway
        :type module_name: str
        :param device_status_provider: Provider of device's current status
        :type device_status_provider: Callable[[str], DeviceStatus]
        :param actuation_handler: Setter of new device actuator values
        :type actuation_handler: Optional[Callable[[str, str, str], None]]
        :param actuator_status_provider: Provider of device's current actuator status
        :type actuator_status_provider: Optional[Callable[[str, str], Tuple[ActuatorState, Union[bool, int, float, str]]]]
        :param configuration_handler: Setter of new device configuration values
        :type configuration_handler: Optional[Callable[[str, Dict[str, Union[bool, int, float, str]]], None]]
        :param configuration_provider: Provider of device's configuration options
        :type configuration_provider: Optional[Callable[[str], Dict[str, Union[int, float, bool, str, Tuple[int, int], Tuple[int, int, int], Tuple[float, float], Tuple[float, float, float], Tuple[str, str], Tuple[str, str, str]]]]]
        :param install_firmware: Handling of firmware installation
        :type install_firmware: Optional[Callable[[str, str], None]]
        :param connectivity_service: Custom connectivity service implementation
        :type connectivity_service: Optional[ConnectivityService]
        :param data_protocol: Custom data protocol implementation
        :type data_protocol: Optional[DataProtocol]
        :param firmware_update_protocol: Custom firmware update protocol implementation
        :type firmware_update_protocol: Optional[FirmwareUpdateProtocol]
        :param registration_protocol: Custom registration protocol implementation
        :type registration_protocol: Optional[RegistrationProtocol]
        :param status_protocol: Custom device status protocol implementation
        :type status_protocol: Optional[StatusProtocol]
        :param outbound_message_queue: Custom persistent storage implementation
        :type outbound_message_queue: Optional[OutboundMessageQueue]
        """
        self.log = logger_factory.get_logger(str(self.__class__.__name__))

        self.host = host
        self.port = port
        self.module_name = module_name

        if not callable(device_status_provider):
            raise RuntimeError(f"{device_status_provider} is not a callable!")
        if len(signature(device_status_provider).parameters) != 1:
            raise RuntimeError(f"{device_status_provider} invalid signature!")
        self.device_status_provider = device_status_provider

        if actuation_handler is not None:
            if not callable(actuation_handler):
                raise RuntimeError(f"{actuation_handler} is not a callable!")
            if len(signature(actuation_handler).parameters) != 3:
                raise RuntimeError(f"{actuation_handler} invalid signature!")
            self.actuation_handler: Optional[
                Callable[[str, str, Union[bool, int, float, str]], None]
            ] = actuation_handler
        else:
            self.actuation_handler = None

        if actuator_status_provider is not None:
            if not callable(actuator_status_provider):
                raise RuntimeError(
                    f"{actuator_status_provider} is not a callable!"
                )
            if len(signature(actuator_status_provider).parameters) != 2:
                raise RuntimeError(
                    f"{actuator_status_provider} invalid signature!"
                )
            self.actuator_status_provider: Optional[
                Callable[
                    [str, str],
                    Tuple[ActuatorState, Union[bool, int, float, str]],
                ]
            ] = actuator_status_provider
        else:
            self.actuator_status_provider = None

        if (
            self.actuation_handler is None
            and self.actuator_status_provider is not None
        ) or (
            self.actuation_handler is not None
            and self.actuator_status_provider is None
        ):
            raise RuntimeError(
                "Provide actuation_handler and actuator_status_provider"
                " to enable actuators on your devices!"
            )

        if configuration_handler is not None:
            if not callable(configuration_handler):
                raise RuntimeError(
                    f"{configuration_handler} is not a callable!"
                )
            if len(signature(configuration_handler).parameters) != 2:
                raise RuntimeError(
                    f"{configuration_handler} invalid signature!"
                )
            self.configuration_handler: Optional[
                Callable[[str, Configuration], None]
            ] = configuration_handler
        else:
            self.configuration_handler = None

        if configuration_provider is not None:
            if not callable(configuration_provider):
                raise RuntimeError(
                    f"{configuration_provider} is not a callable!"
                )
            if len(signature(configuration_provider).parameters) != 1:
                raise RuntimeError(
                    f"{configuration_provider} invalid signature!"
                )
            self.configuration_provider: Optional[
                Callable[[str], Configuration]
            ] = configuration_provider
        else:
            self.configuration_provider = None

        if (
            self.configuration_handler is None
            and self.configuration_provider is not None
        ) or (
            self.configuration_handler is not None
            and self.configuration_provider is None
        ):
            raise RuntimeError(
                "Provide configuration_handler and configuration_provider"
                " to enable configuration options on your devices!"
            )

        if firmware_handler is not None:
            if not isinstance(firmware_handler, FirmwareHandler):
                raise RuntimeError(
                    f"{firmware_handler} isn't an instance of FirmwareHandler!"
                )
            self.firmware_handler: Optional[FirmwareHandler] = firmware_handler
            self.firmware_handler.on_install_success = (
                self._on_install_success  # type: ignore
            )
            self.firmware_handler.on_install_fail = (
                self._on_install_fail  # type: ignore
            )
        else:
            self.firmware_handler = None

        if data_protocol is not None:
            if not isinstance(data_protocol, DataProtocol):
                raise RuntimeError(
                    f"{data_protocol} is not a valid instance of DataProtocol!"
                )
            self.data_protocol = data_protocol
        else:
            self.data_protocol = JsonDataProtocol()

        if firmware_update_protocol is not None:
            if not isinstance(
                firmware_update_protocol, FirmwareUpdateProtocol
            ):
                raise RuntimeError(
                    f"{firmware_update_protocol} is not a valid instance of"
                    " FirmwareUpdateProtocol!"
                )
            self.firmware_update_protocol = firmware_update_protocol
        else:
            self.firmware_update_protocol = JsonFirmwareUpdateProtocol()

        if status_protocol is not None:
            if not isinstance(status_protocol, StatusProtocol):
                raise RuntimeError(
                    f"{status_protocol} is not a valid instance of "
                    "StatusProtocol!"
                )
            self.status_protocol = status_protocol
        else:
            self.status_protocol = JsonStatusProtocol()

        if registration_protocol is not None:
            if not isinstance(registration_protocol, RegistrationProtocol):
                raise RuntimeError(
                    f"{registration_protocol} is not a valid instance of "
                    "RegistrationProtocol!"
                )
            self.registration_protocol = registration_protocol
        else:
            self.registration_protocol = JsonRegistrationProtocol()

        if outbound_message_queue is not None:
            if not isinstance(outbound_message_queue, OutboundMessageQueue):
                raise RuntimeError(
                    f"{outbound_message_queue} is not a valid instance of "
                    "OutboundMessageQueue!"
                )
            self.outbound_message_queue = outbound_message_queue
        else:
            self.outbound_message_queue = OutboundMessageDeque()

        self.devices: List[Device] = []

        last_will_message = self.status_protocol.make_last_will_message(
            [device.key for device in self.devices]
        )

        if connectivity_service is not None:
            if not isinstance(connectivity_service, ConnectivityService):
                raise RuntimeError(
                    f"{connectivity_service} is not a valid instance of "
                    "ConnectivityService!"
                )
            self.connectivity_service = connectivity_service
        else:
            self.connectivity_service = MQTTConnectivityService(
                host, port, module_name, 0, last_will_message, []
            )

        self.connectivity_service.set_inbound_message_listener(
            self._on_inbound_message
        )

        self.log.debug(self.__repr__())

    @recursive_repr()
    def __repr__(self) -> str:
        """
        Make string representation or Wolk.

        :returns: representation
        :rtype: str
        """
        return (
            f"Wolk(host='{self.host}', "
            f"port='{self.port}', "
            f"module_name='{self.module_name}', "
            f"device_status_provider='{self.device_status_provider}', "
            f"actuation_handler='{self.actuation_handler}', "
            f"actuator_status_provider='{self.actuator_status_provider}', "
            f"configuration_handler='{self.configuration_handler}', "
            f"configuration_provider='{self.configuration_provider}', "
            f"firmware_handler='{self.firmware_handler}', "
            f"data_protocol='{self.data_protocol}', "
            f"firmware_update_protocol='{self.firmware_update_protocol}', "
            f"status_protocol='{self.status_protocol}', "
            f"registration_protocol='{self.registration_protocol}', "
            f"outbound_message_queue='{self.outbound_message_queue}', "
            f"connectivity_service='{self.connectivity_service}', "
            f"devices='{self.devices}')"
        )

    def _on_install_success(self, device_key: str) -> None:
        """
        Handle firmware installation message from firmware_handler.

        :param device_key: Device that completed firmware update
        :type device_key: str
        """
        self.log.info(
            f"Received firmware installation success for device '{device_key}'"
        )
        status = FirmwareUpdateStatus(FirmwareUpdateState.COMPLETED)
        message = self.firmware_update_protocol.make_update_message(
            device_key, status
        )
        if not self.connectivity_service.publish(message):
            if not self.outbound_message_queue.put(message):
                self.log.error(
                    "Failed to publish or store "
                    f"firmware version message {message}"
                )
                return
        if self.firmware_handler:
            version = self.firmware_handler.get_firmware_version(device_key)
            if not version:
                self.log.error(
                    "Did not get firmware version for "
                    f"device '{device_key}'"
                )
                return
            message = self.firmware_update_protocol.make_version_message(
                device_key, version
            )
            if not self.connectivity_service.publish(message):
                if not self.outbound_message_queue.put(message):
                    self.log.error(
                        "Failed to publish or store "
                        f"firmware version message {message}"
                    )

    def _on_install_fail(
        self, device_key: str, status: FirmwareUpdateStatus
    ) -> None:
        """
        Handle firmware installation failiure from firmware_handler.

        :param device_key: Device that reported firmware installation error
        :type device_key: str
        :param status: Firware update status information
        :type status: FirmwareUpdateStatus
        """
        self.log.info(
            "Received firmware installation status "
            f"message '{status}' for device '{device_key}'"
        )
        if not isinstance(status, FirmwareUpdateStatus):
            self.log.error(  # type: ignore
                f"Received status {status} is not "
                "an instance of FirmwareUpdateStatus!"
            )
            return

        message = self.firmware_update_protocol.make_update_message(
            device_key, status
        )
        if not self.connectivity_service.publish(message):
            if not self.outbound_message_queue.put(message):
                self.log.error(
                    "Failed to publish or store "
                    f"firmware version message {message}"
                )

    def _on_inbound_message(self, message: Message) -> None:
        """
        Handle messages received from WolkGateway.

        :param message: Message received
        :type message: Message
        """
        self.log.debug(f"Received message: {message}")

        if self.data_protocol.is_actuator_set_message(message):

            if not (self.actuation_handler and self.actuator_status_provider):
                self.log.warning(
                    f"Received actuation message {message} , but no "
                    "actuation handler and actuator status provider present"
                )
                return

            self.log.info(f"Received actuator set command: {message}")
            device_key = self.data_protocol.extract_key_from_message(message)
            device_status = self.device_status_provider(device_key)
            if device_status not in [
                DeviceStatus.CONNECTED,
                DeviceStatus.SLEEP,
            ]:
                self.log.warning(
                    f"Device '{device_key}' returned '{device_status.value}' "
                    "status, not forwarding command"
                )
                self.publish_device_status(device_key)
                return

            actuator_command_set = self.data_protocol.make_actuator_command(
                message
            )
            self.actuation_handler(
                device_key,
                actuator_command_set.reference,
                actuator_command_set.value,  # type: ignore
            )
            try:
                self.publish_actuator_status(
                    device_key, actuator_command_set.reference
                )
                return
            except RuntimeError as e:
                self.log.error(
                    "Error occurred during handing"
                    f" inbound actuation message {message} : {e}"
                )

        elif self.data_protocol.is_actuator_get_message(message):

            if not (self.actuation_handler and self.actuator_status_provider):
                self.log.warning(
                    f"Received actuation message {message} , but no "
                    "actuation handler and actuator status provider present"
                )
                return

            self.log.info(f"Received actuator get command: {message}")
            device_key = self.data_protocol.extract_key_from_message(message)
            device_status = self.device_status_provider(device_key)
            if device_status not in [
                DeviceStatus.CONNECTED,
                DeviceStatus.SLEEP,
            ]:
                self.log.warning(
                    f"Device '{device_key}' returned '{device_status.value}' "
                    "status, not forwarding command"
                )
                self.publish_device_status(device_key)
                return

            actuator_command_get = self.data_protocol.make_actuator_command(
                message
            )
            try:
                self.publish_actuator_status(
                    device_key, actuator_command_get.reference
                )
            except RuntimeError as e:
                self.log.error(
                    "Error occurred during handing "
                    f"inbound actuation message {message} : {e}"
                )

        elif self.data_protocol.is_configuration_set_message(message):

            if not (
                self.configuration_handler and self.configuration_provider
            ):
                self.log.warning(
                    f"Received configuration message {message} , but no "
                    "configuration handler and configuration provider present"
                )
                return

            self.log.info(f"Received configuration set command: {message}")
            device_key = self.data_protocol.extract_key_from_message(message)
            device_status = self.device_status_provider(device_key)
            if device_status not in [
                DeviceStatus.CONNECTED,
                DeviceStatus.SLEEP,
            ]:
                self.log.warning(
                    f"Device '{device_key}' returned '{device_status.value}' "
                    "status, not forwarding command"
                )
                self.publish_device_status(device_key)
                return

            config_set = self.data_protocol.make_configuration_command(message)
            if config_set.value is not None:
                self.configuration_handler(device_key, config_set.value)
                try:
                    self.publish_configuration(device_key)
                except RuntimeError as e:
                    self.log.error(
                        "Error occurred during handling "
                        f"inbound configuration message {message} : {e}"
                    )
                    return
            else:
                self.log.warning(
                    "Received malformed configuration message: "
                    f"{message}\nParser yielded: {config_set}"
                )
                return

        elif self.data_protocol.is_configuration_get_message(message):

            if not (
                self.configuration_handler and self.configuration_provider
            ):
                self.log.warning(
                    f"Received configuration message {message} , but no "
                    "configuration handler and configuration provider present"
                )
                return

            self.log.info(f"Received configuration get command: {message}")
            device_key = self.data_protocol.extract_key_from_message(message)
            device_status = self.device_status_provider(device_key)
            if device_status not in [
                DeviceStatus.CONNECTED,
                DeviceStatus.SLEEP,
            ]:
                self.log.warning(
                    f"Device '{device_key}' returned '{device_status.value}' "
                    "status, not forwarding command"
                )
                self.publish_device_status(device_key)
                return

            try:
                self.publish_configuration(device_key)
            except RuntimeError as e:
                self.log.error(
                    "Error occurred during handling "
                    f"inbound configuration message {message} : {e}"
                )
                return

        elif self.registration_protocol.is_registration_response_message(
            message
        ):

            response = self.registration_protocol.make_registration_response(
                message
            )
            if response.key not in [device.key for device in self.devices]:
                self.log.warning(
                    f"Received unexpected registration response: {message}"
                )
                return

            self.log.info(f"Received registration response: {response}")

            for device in self.devices:
                if device.key == response.key:
                    registered_device = device
                    break

            device_status = self.device_status_provider(registered_device.key)
            if device_status not in [
                DeviceStatus.CONNECTED,
                DeviceStatus.SLEEP,
            ]:
                self.log.warning(
                    f"Device '{registered_device.key}' returned "
                    f"'{device_status.value}' "
                    "status, not getting device data"
                )
                self.publish_device_status(device_key)
                return

            if registered_device.get_actuator_references():
                for reference in registered_device.get_actuator_references():
                    try:
                        self.publish_actuator_status(
                            registered_device.key, reference
                        )
                    except RuntimeError as e:
                        self.log.error(
                            "Error occurred when sending actuator status "
                            f"for device {registered_device.key} with "
                            f"reference {reference} : {e}"
                        )

            if registered_device.has_configurations():
                try:
                    self.publish_configuration(registered_device.key)
                except RuntimeError as e:
                    self.log.error(
                        "Error occurred when sending configuration "
                        f"for device {registered_device.key} : {e}"
                    )

            if registered_device.supports_firmware_update():
                if self.firmware_handler is not None:
                    version = self.firmware_handler.get_firmware_version(
                        registered_device.key
                    )
                    if not version:
                        self.log.error(
                            "Did not get firmware version for "
                            f"device '{registered_device.key}'"
                        )
                        return

                    msg = self.firmware_update_protocol.make_version_message(
                        registered_device.key, version
                    )
                    if not self.connectivity_service.publish(msg):
                        if not self.outbound_message_queue.put(msg):
                            self.log.error(
                                "Failed to publish or store "
                                f"firmware version message {msg}"
                            )
                            return

        elif self.status_protocol.is_device_status_request_message(message):

            self.log.info(f"Received device status request: {message}")
            device_key = self.status_protocol.extract_key_from_message(message)
            status = self.device_status_provider(device_key)
            if not status:
                self.log.error(
                    "Device status provider didn't return a "
                    f"status for device '{device_key}'"
                )
                return
            message = self.status_protocol.make_device_status_response_message(
                status, device_key
            )
            if not self.connectivity_service.publish(message):
                if not self.outbound_message_queue.put(message):
                    self.log.error(
                        "Failed to publish or store "
                        f"device status message {message}"
                    )

        elif self.firmware_update_protocol.is_firmware_install_command(
            message
        ):

            if self.firmware_handler is None:
                self.log.warning(
                    "No firmware handler, ignoring message: " f"{message}"
                )
                return

            key = self.firmware_update_protocol.extract_key_from_message(
                message
            )
            device_status = self.device_status_provider(key)
            if device_status not in [
                DeviceStatus.CONNECTED,
                DeviceStatus.SLEEP,
            ]:
                self.log.warning(
                    f"Device '{key}' returned '{device_status.value}' "
                    "status, not forwarding command"
                )
                self.publish_device_status(key)
                return

            path = self.firmware_update_protocol.make_firmware_file_path(
                message
            )
            self.log.info(
                "Received firmware installation command "
                f"for device '{key}' with file path: {path}"
            )
            firmware_status = FirmwareUpdateStatus(
                FirmwareUpdateState.INSTALLATION
            )
            update_message = self.firmware_update_protocol.make_update_message(
                key, firmware_status
            )
            if not self.connectivity_service.publish(update_message):
                if not self.outbound_message_queue.put(update_message):
                    self.log.error(
                        "Failed to publish or store "
                        f"firmware update status message {update_message}"
                    )

            self.firmware_handler.install_firmware(key, path)

        elif self.firmware_update_protocol.is_firmware_abort_command(message):

            if self.firmware_handler is None:
                self.log.warning(
                    "No firmware handler, ignoring message: " f"{message}"
                )
                return

            key = self.firmware_update_protocol.extract_key_from_message(
                message
            )
            self.log.info(
                "Received firmware installation abort command for device {key}"
            )
            self.firmware_handler.abort_installation(key)

    def add_sensor_reading(
        self,
        device_key: str,
        reference: str,
        value: Union[
            bool,
            int,
            float,
            str,
            Tuple[int, int],
            Tuple[int, int, int],
            Tuple[float, float],
            Tuple[float, float, float],
            Tuple[str, str],
            Tuple[str, str, str],
        ],
        timestamp: Optional[int] = None,
    ) -> None:
        """
        Serialize sensor reading and put into storage.

        Storing readings without Unix timestamp will result
        in all sent messages being treated as live readings and
        will be assigned a timestamp upon reception, so for a valid
        history add timestamps to readings via ``int(round(time.time() * 1000))``

        :param device_key: Device on which the sensor reading occurred
        :type device_key: str
        :param reference: Sensor reference (unique per device)
        :type reference: str
        :param value: Value(s) that the reading yielded
        :type value: Union[bool,int,float,str,Tuple[int, int],Tuple[int, int, int],Tuple[float, float],Tuple[float, float, float],Tuple[str, str],Tuple[str, str, str],]
        :param timestamp: Unix time
        :type timestamp: Optional[int]

        :raises RuntimeError: Unable to place in storage
        """
        self.log.debug(
            f"Add sensor reading: {device_key} , "
            f"{reference} , {value} , {timestamp}"
        )
        reading = SensorReading(reference, value, timestamp)
        message = self.data_protocol.make_sensor_reading_message(
            device_key, reading
        )
        if not self.outbound_message_queue.put(message):
            raise RuntimeError(f"Unable to store message: {message}")

    def add_sensor_readings(
        self,
        device_key: str,
        readings: Dict[
            str,
            Union[
                bool,
                int,
                float,
                str,
                Tuple[int, int],
                Tuple[int, int, int],
                Tuple[float, float],
                Tuple[float, float, float],
                Tuple[str, str],
                Tuple[str, str, str],
            ],
        ],
        timestamp: Optional[int] = None,
    ) -> None:
        """
        Serialize multiple sensor readings and put into storage.

        Storing readings without Unix timestamp will result
        in all sent messages being treated as live readings and
        will be assigned a timestamp upon reception, so for a valid
        history add timestamps to readings via ``int(round(time.time() * 1000))``

        :param device_key: Device on which the sensor reading occurred
        :type device_key: str
        :param readings: dictionary in sensor_reference:value format
        :type readings:  Dict[str,Union[bool,int,float,str,Tuple[int, int],Tuple[int, int, int],Tuple[float, float],Tuple[float, float, float],Tuple[str, str],Tuple[str, str, str],]
        :param timestamp: Unix time
        :type timestamp: Optional[int]

        :raises RuntimeError: Unable to place in storage
        """
        self.log.debug(
            f"Add sensor readings: {device_key} , " f"{readings} , {timestamp}"
        )
        sensor_readings = []
        for reference, value in readings.items():
            sensor_readings.append(SensorReading(reference, value))
        message = self.data_protocol.make_sensor_readings_message(
            device_key, sensor_readings, timestamp
        )
        if not self.outbound_message_queue.put(message):
            raise RuntimeError(f"Unable to store message: {message}")

    def add_alarm(
        self,
        device_key: str,
        reference: str,
        active: bool,
        timestamp: Optional[int] = None,
    ) -> None:
        """
        Serialize alarm event and put into storage.

        Storing alarms without Unix timestamp will result
        in all sent messages being treated as live and
        will be assigned a timestamp upon reception, so for a valid
        history add timestamps to alarms via ``int(round(time.time() * 1000))``

        :param device_key: Device on which the sensor reading occurred
        :type device_key: str
        :param reference: Alarm reference (unique per device)
        :type reference: str
        :param value: Current state of alarm
        :type active: bool
        :param timestamp: Unix time
        :type timestamp: Optional[int]

        :raises RuntimeError: Unable to place in storage
        """
        self.log.debug(
            f"Add alarm: {device_key} , {reference} , {active} , {timestamp}"
        )
        alarm = Alarm(reference, active, timestamp)
        message = self.data_protocol.make_alarm_message(device_key, alarm)
        if not self.outbound_message_queue.put(message):
            raise RuntimeError(f"Unable to store message: {message}")

    def publish_actuator_status(
        self,
        device_key: str,
        reference: str,
        state: Optional[ActuatorState] = None,
        value: Optional[Union[bool, int, float, str]] = None,
    ) -> None:
        """
        Publish device actuator status to WolkGateway.

        Getting the actuator status is achieved by calling the user's
        implementation of ``actuator_status_provider`` or optionally an
        actuator status can be published explicitly
        by providing ``ActuatorState`` as ``state`` parameter and the
        current actuator value via ``value`` parameter

        If message is unable to be sent, it will be placed in storage.

        If no ``actuator_status_provider`` is present, will raise exception.

        :param device_key: Device on which the sensor reading occurred
        :type device_key: str
        :param reference: Alarm reference (unique per device)
        :type reference: str
        :param state: Current actuator state for explicitly publishing status
        :type state: Optional[ActuatorState]
        :param value: Current actuator value for explicitly publishing status
        :type value: Optional[Union[bool, int, float, str]]

        :raises ValueError: Provided state is not an instance of ActuatorState
        :raises RuntimeError: Unable to place in storage or no status provider
        """
        self.log.debug(f"Publish actuator status: {device_key} , {reference}")
        if not (self.actuator_status_provider and self.actuation_handler):
            raise RuntimeError(
                "Unable to publish actuator status because "
                "actuator_status_provider and actuation_handler "
                "were not provided!"
            )
        if state is not None:
            if not isinstance(state, ActuatorState):
                raise ValueError(
                    f"{state} is not an instance of ActuatorState"
                )
            if state != ActuatorState.ERROR and value is None:
                raise ValueError(
                    f"Value must be provided for actuator state"
                    f" '{state.value}' !"
                )
        else:
            state, value = self.actuator_status_provider(device_key, reference)
            self.log.debug(
                f"Actuator status provider returned: {state} {value}"
            )

            if state is None:
                raise RuntimeError(
                    f"{self.actuator_status_provider} did not return anything"
                    f" for device '{device_key}' with reference '{reference}'"
                )

            if not isinstance(state, ActuatorState):
                raise RuntimeError(
                    f"{state} is not a member of ActuatorState!"
                )

        status = ActuatorStatus(reference, state, value)
        message = self.data_protocol.make_actuator_status_message(
            device_key, status
        )
        if self.connectivity_service.connected():
            if not self.connectivity_service.publish(message):
                if not self.outbound_message_queue.put(message):
                    raise RuntimeError(
                        f"Unable to publish and failed "
                        f"to store message: {message}"
                    )
        else:
            self.log.warning(
                "Not connected, unable to publish "
                f"actuator status message {message}"
            )
            if not self.outbound_message_queue.put(message):
                raise RuntimeError(f"Unable to store message: {message}")

    def publish_device_status(
        self, device_key: str, status: Optional[DeviceStatus] = None
    ) -> None:
        """
        Publish current device status to WolkGateway.

        Getting the current device status is achieved by calling the user's
        provided ``device_status_provider`` or a device status can be published
        explicitly by passing a ``DeviceStatus`` as the ``status`` parameter.

        :param device_key: Device to which the status belongs to
        :type device_key: str
        :param status: Current device status
        :type status: Optional[DeviceStatus]

        :raises ValueError: status is not of ``DeviceStatus``
        :raises RuntimeError: Failed to publish and store message
        """
        self.log.debug(f"Publish device status for {device_key}")

        if status is not None:
            if not isinstance(status, DeviceStatus):
                raise ValueError(
                    f"{status} is not an instance of DeviceStatus"
                )

        else:
            status = self.device_status_provider(device_key)
            if not isinstance(status, DeviceStatus):
                raise ValueError(
                    f"{status} is not an instance of DeviceStatus"
                )

        message = self.status_protocol.make_device_status_update_message(
            status, device_key
        )

        if self.connectivity_service.connected():
            if not self.connectivity_service.publish(message):
                if not self.outbound_message_queue.put(message):
                    raise RuntimeError(
                        f"Unable to publish and failed "
                        f"to store message: {message}"
                    )
        else:
            self.log.warning(
                "Not connected, unable to publish "
                f"device status message {message}"
            )
            if not self.outbound_message_queue.put(message):
                raise RuntimeError(f"Unable to store message: {message}")

    def publish_configuration(self, device_key: str) -> None:
        """
        Publish device configuration options to WolkGateway.

        If message is unable to be sent, it will be placed in storage.

        Getting the current configuration is achieved by calling the user's
        implementation of ``configuration_provider``.

        If no ``configuration_provider`` is present, will raise exception.

        :param device_key: Device to which the configuration belongs to
        :type device_key: str

        :raises RuntimeError: No configuration provider present or no data returned
        """
        self.log.debug(f"Publish configuration: {device_key}")
        if not (self.configuration_handler and self.configuration_provider):
            raise RuntimeError(
                "Unable to publish configuration because "
                "configuration_provider and configuration_handler "
                "were not provided!"
            )

        configuration = self.configuration_provider(device_key)

        self.log.debug(f"Configuration provider returned: {configuration}")

        if configuration is None:
            raise RuntimeError(
                f"{self.configuration_provider} did not return"
                f"anything for device '{device_key}'"
            )

        message = self.data_protocol.make_configuration_message(
            device_key, configuration
        )
        if self.connectivity_service.connected():
            if not self.connectivity_service.publish(message):
                if not self.outbound_message_queue.put(message):
                    raise RuntimeError(
                        f"Unable to publish and failed "
                        f"to store message: {message}"
                    )
        else:
            self.log.warning(
                "Not connected, unable to publish "
                f"configuration status message {message}"
            )
            if not self.outbound_message_queue.put(message):
                raise RuntimeError(f"Unable to store message: {message}")

    def add_device(self, device: Device) -> None:
        """
        Add device to module.

        Will attempt to send a registration request and
        update list of subscribed topics.

        :param device: Device to be added to module
        :type device: Device

        :raises RuntimeError: Unable to store message
        :raises ValueError: Invalid device given
        """
        self.log.debug(f"Add device: {device}")
        if not isinstance(device, Device):
            raise ValueError(
                "Given device is not an instance of Device class!"
            )
        device_keys = [device.key for device in self.devices]
        if device.key in device_keys:
            self.log.error(f"Device with key '{device.key}' was already added")
            return

        if device.get_actuator_references():
            if not (self.actuation_handler and self.actuator_status_provider):
                self.log.error(
                    f"Can not add device '{device.key}' with actuators "
                    "without having an actuation handler and "
                    "actuator status provider"
                )
                return

        if device.has_configurations():
            if not (
                self.configuration_handler and self.configuration_provider
            ):
                self.log.error(
                    f"Can not add device '{device.key}' with "
                    "configuration options without having a "
                    "configuration handler and configuration provider"
                )
                return

        if device.supports_firmware_update():
            if not self.firmware_handler:
                self.log.error(
                    f"Can not add device '{device.key}' with "
                    "firmware update support without having a "
                    "firmware handler"
                )
                return

        self.devices.append(device)

        device_topics = []
        device_topics.extend(
            self.data_protocol.get_inbound_topics_for_device(device.key)
        )
        device_topics.extend(
            self.registration_protocol.get_inbound_topics_for_device(
                device.key
            )
        )
        device_topics.extend(
            self.firmware_update_protocol.get_inbound_topics_for_device(
                device.key
            )
        )
        device_topics.extend(
            self.status_protocol.get_inbound_topics_for_device(device.key)
        )

        self.connectivity_service.add_subscription_topics(device_topics)

        self.connectivity_service.set_lastwill_message(
            self.status_protocol.make_last_will_message(
                [device.key for device in self.devices]
            )
        )

        registration_request = DeviceRegistrationRequest(
            device.name, device.key, device.template
        )

        message = self.registration_protocol.make_registration_message(
            registration_request
        )

        if not self.connectivity_service.connected():
            if not self.outbound_message_queue.put(message):
                raise RuntimeError(f"Unable to store message: {message}")
        else:
            try:
                if not self.connectivity_service.reconnect():
                    self.log.error("Failed to reconnect")
            except RuntimeError as e:
                self.log.error(f"Failed to reconnect: {e}")
                if not self.outbound_message_queue.put(message):
                    raise RuntimeError(f"Unable to store message: {message}")
            if not self.connectivity_service.publish(message):
                if not self.outbound_message_queue.put(message):
                    raise RuntimeError(f"Unable to store message: {message}")

    def remove_device(self, device_key: str) -> None:
        """
        Remove device from module.

        Removes device for subscription topics and lastwill message.

        :param device_key: Device identifier
        :type device_key: str
        """
        self.log.debug(f"Removing device: {device_key}")
        if device_key not in [device.key for device in self.devices]:
            self.log.info(f"Device with key '{device_key}' was not stored")
            return

        for device in self.devices:
            if device_key == device.key:
                self.devices.remove(device)
                break

        self.connectivity_service.remove_topics_for_device(device_key)

        self.connectivity_service.set_lastwill_message(
            self.status_protocol.make_last_will_message(
                [device.key for device in self.devices]
            )
        )

        if self.connectivity_service.connected():
            try:
                self.connectivity_service.reconnect()
            except RuntimeError as e:
                self.log.error(f"Failed to reconnect: {e}")

    def publish(self, device_key: Optional[str] = None) -> None:
        """
        Publish stored messages to WolkGateway.

        If device_key parameter is provided, will publish messages only
        for that specific device.

        :param device_key: Device for which to publish stored messages
        :type device_key: Optional[str]
        """
        if device_key:
            self.log.debug(f"Publishing messages for {device_key}")
        else:
            self.log.debug("Publishing all stored messages")

        if self.outbound_message_queue.queue_size() == 0:
            self.log.info("No messages to publish")
            return

        if not self.connectivity_service.connected():
            self.log.warning("Not connected, unable to publish")
            return

        if device_key is None:
            while self.outbound_message_queue.queue_size() > 0:
                message = self.outbound_message_queue.get()
                if message is None:
                    return
                if not self.connectivity_service.publish(message):
                    self.log.error(f"Failed to publish {message}")
                    sleep(0.2)
                    self.log.info(f"Retrying publish {message}")
                    if not self.connectivity_service.publish(message):
                        self.log.error(f"Failed to publish {message}")
                        return
                self.outbound_message_queue.remove(message)
        else:
            messages = self.outbound_message_queue.get_messages_for_device(
                device_key
            )
            if len(messages) == 0:
                self.log.warning(f"No messages stored for {device_key}")
                return
            for message in messages:
                if not self.connectivity_service.publish(message):
                    self.log.error(f"Failed to publish {message}")
                    sleep(0.2)
                    self.log.info(f"Retrying publish {message}")
                    if not self.connectivity_service.publish(message):
                        self.log.error(f"Failed to publish {message}")
                        return
                self.outbound_message_queue.remove(message)

    def connect(self) -> None:
        """
        Establish connection with WolkGateway.

        Will attempt to publish actuator statuses, configuration options,
        and current firmware version for all added devices.

        :raises RuntimeError: Error publishing actuator status or configuration
        """
        self.log.debug("Connecting to WolkGateway")
        if self.connectivity_service.connected():
            self.log.info("Already connected")
        else:
            try:
                if not self.connectivity_service.connect():
                    self.log.error("Failed to connect")
            except RuntimeError as e:
                self.log.error(f"Failed to connect: {e}")
                return

        if self.connectivity_service.connected():
            self.log.info("Connection to gateway established")
            for device in self.devices:
                try:
                    device_status = self.device_status_provider(device.key)
                    if device_status not in [
                        DeviceStatus.CONNECTED,
                        DeviceStatus.SLEEP,
                    ]:
                        self.log.warning(
                            f"Device '{device.key}' returned "
                            f"'{device_status.value}' "
                            "status, not getting device data"
                        )
                        self.publish_device_status(device.key)
                        continue

                    self.publish_device_status(device.key)

                except (ValueError, RuntimeError) as e:
                    raise e

                for reference in device.get_actuator_references():
                    try:
                        self.publish_actuator_status(device.key, reference)
                    except RuntimeError as e:
                        raise e

                if device.has_configurations():
                    try:
                        self.publish_configuration(device.key)
                    except RuntimeError as e:
                        raise e

                if device.supports_firmware_update():
                    if self.firmware_handler is None:
                        self.log.warning(
                            "Module does not support firmware update, "
                            "not forwarding firmware version for device "
                            f"with key '{device.key}'"
                        )
                        return
                    version = self.firmware_handler.get_firmware_version(
                        device.key
                    )
                    if not version:
                        self.log.error(
                            "Did not get firmware version for "
                            f"device '{device.key}'"
                        )
                        continue
                    msg = self.firmware_update_protocol.make_version_message(
                        device.key, version
                    )
                    if not self.connectivity_service.publish(msg):
                        if not self.outbound_message_queue.put(msg):
                            raise RuntimeError(
                                "Failed to publish or store "
                                f"firmware version message {msg}"
                            )

    def disconnect(self) -> None:
        """Terminate connection with WolkGateway."""
        self.log.debug("Disconnecting from WolkGateway")
        if not self.connectivity_service.connected():
            self.log.debug("Not connected")
            return
        else:
            self.connectivity_service.disconnect()



"""
Data types for creating generic reading types.

Used to visualise data on WolkAbout IoT Platform.
"""
#   Copyright 2019 WolkAbout Technology s.r.o.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
from enum import Enum
from enum import unique


@unique
class DataType(Enum):
    """
    Use to create a generic reading type.

    :ivar BOOLEAN: Generic boolean reading type
    :vartype BOOLEAN: int
    :ivar NUMERIC: Generic numeric reading type
    :vartype NUMERIC: int
    :ivar STRING: Generic string reading type
    :vartype STRING: int
    """

    NUMERIC = 0
    BOOLEAN = 1
    STRING = 2



"Tests for OutboundMessageDeque."
#   Copyright 2019 WolkAbout Technology s.r.o.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
import sys
import unittest

sys.path.append("..")  # noqa

from wolk_gateway_module.outbound_message_deque import OutboundMessageDeque
from wolk_gateway_module.model.message import Message


class OutboundMessageDequeTests(unittest.TestCase):
    """OutboundMessageDeque Tests."""

    def test_put(self):
        """Test put message into storage."""
        deque = OutboundMessageDeque()

        self.assertTrue(deque.put("test"))
        self.assertEqual(deque.queue_size(), 1)

    def test_get_something(self):
        """Test get from empty storage."""
        deque = OutboundMessageDeque()
        value = "test"
        deque.put(value)

        self.assertEqual(value, deque.get())

    def test_get_nothing(self):
        """Test get message from storage."""
        deque = OutboundMessageDeque()
        deque.put(None)

        self.assertEqual(None, deque.get())

    def test_queue_size(self):
        """Test getting queue storage size."""
        deque = OutboundMessageDeque()
        deque.put(1)
        deque.put(2)
        deque.put(3)

        self.assertEqual(3, deque.queue_size())

    def test_remove(self):
        """Test removing from queue."""
        deque = OutboundMessageDeque()
        deque.put(1)
        deque.put(2)
        deque.put(3)

        deque.remove(2)

        self.assertEqual(2, deque.queue_size())

    def test_get_no_messages_for_device(self):
        """Test getting no messages for specific device from empty queue."""
        deque = OutboundMessageDeque()
        device_key = "some_key"

        self.assertFalse(deque.get_messages_for_device(device_key))

    def test_get_no_matching_messages_for_device(self):
        """Test getting no messages for specific device from queue."""
        deque = OutboundMessageDeque()
        device_key = "some_key"

        deque.put(Message("some_other_key"))
        deque.put(Message("another_key"))
        deque.put(Message("different_key"))

        self.assertFalse(deque.get_messages_for_device(device_key))

    def test_get_messages_for_device(self):
        """Test getting some messages for specific device from queue."""
        deque = OutboundMessageDeque()
        device_key = "some_key"

        deque.put(Message("some_key/2"))
        deque.put(Message("another_key"))
        deque.put(Message("some_key/1"))

        self.assertEqual(
            [Message("some_key/2"), Message("some_key/1")],
            deque.get_messages_for_device(device_key),
        )


if __name__ == "__main__":
    unittest.main()



"""Abstract base class for handling device firmware update."""
#   Copyright 2019 WolkAbout Technology s.r.o.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
from abc import ABC
from abc import abstractmethod


class FirmwareHandler(ABC):
    """
    Handle firmware installation and abort commands, and report version.

    Once an object of this class is passed to a Wolk object,
    it will set callback methods `on_install_success` and
    `on_install_fail` used for reporting the result of
    the firmware update process. Use these callbacks in `install_firmware`
    and `abort_installation` methods.

    :ivar on_install_fail: Installation failure callback method
    :vartype on_install_fail: Callable[[str, FirmwareUpdateStatus], None]
    :ivar on_install_success: Installation successful callback method
    :vartype on_install_success: Callable[[str], None]
    """

    on_install_success = None
    on_install_fail = None

    @abstractmethod
    def install_firmware(
        self, device_key: str, firmware_file_path: str
    ) -> None:
        """
        Handle the installation of the firmware file.

        Call ``self.on_install_success(device_key)`` to report success.
        Reporting success will also get new firmware version.

        If installation fails, call ``self.on_install_fail(device_key, status)``
        where:

        .. code-block:: python

                status = FirmwareUpdateStatus(
                    FirmwareUpdateState.ERROR,
                    FirmwareUpdateErrorCode.INSTALLATION_FAILED
                )

        or use other values from ``FirmwareUpdateErrorCode`` if they fit better.

        :param device_key: Device for which the firmware command is intended
        :type device_key: str
        :param firmware_file_path: Path where the firmware file is located
        :type firmware_file_path: str
        """
        pass

    @abstractmethod
    def abort_installation(self, device_key: str) -> None:
        """
        Attempt to abort the firmware installation process for device.

        Call ``self.on_install_fail(device_key, status)`` to report if
        the installation process was able to be aborted with
        ``status = FirmwareUpdateStatus(FirmwareUpdateState.ABORTED)``.
        If unable to stop the installation process, no action is required.

        :param device_key: Device for which to abort installation
        :type device_key: str
        """
        pass

    @abstractmethod
    def get_firmware_version(self, device_key: str) -> str:
        """
        Return device's current firmware version.

        :param device_key: Device identifier
        :type device_key: str
        :returns: version
        :rtype: str
        """
        pass



"""Handling of messages related to device firmware update."""
#   Copyright 2019 WolkAbout Technology s.r.o.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
import json

from wolk_gateway_module.logger_factory import logger_factory
from wolk_gateway_module.model.firmware_update_status import (
    FirmwareUpdateStatus,
)
from wolk_gateway_module.model.message import Message
from wolk_gateway_module.protocol.firmware_update_protocol import (
    FirmwareUpdateProtocol,
)


class JsonFirmwareUpdateProtocol(FirmwareUpdateProtocol):
    """Parse inbound messages and serialize outbound firmware messages."""

    DEVICE_PATH_PREFIX = "d/"
    FIRMWARE_UPDATE_INSTALL_TOPIC_ROOT = "p2d/firmware_update_install/"
    FIRMWARE_UPDATE_ABORT_TOPIC_ROOT = "p2d/firmware_update_abort/"
    FIRMWARE_UPDATE_STATUS_TOPIC_ROOT = "d2p/firmware_update_status/"
    FIRMWARE_VERSION_UPDATE_TOPIC_ROOT = "d2p/firmware_version_update/"

    def __init__(self) -> None:
        """Create object."""
        self.log = logger_factory.get_logger(str(self.__class__.__name__))

    def __repr__(self) -> str:
        """
        Make string representation of JsonFirmwareUpdateProtocol.

        :returns: representation
        :rtype: str
        """
        return "JsonFirmwareUpdateProtocol()"

    def get_inbound_topics_for_device(self, device_key: str) -> list:
        """
        Return list of inbound topics for given device key.

        :param device_key: Device key for which to create topics
        :type device_key: str

        :returns: inbound_topics
        :rtype: list
        """
        inbound_topics = [
            self.FIRMWARE_UPDATE_INSTALL_TOPIC_ROOT
            + self.DEVICE_PATH_PREFIX
            + device_key,
            self.FIRMWARE_UPDATE_ABORT_TOPIC_ROOT
            + self.DEVICE_PATH_PREFIX
            + device_key,
        ]
        self.log.debug(f"Inbound topics for {device_key} : {inbound_topics}")
        return inbound_topics

    def make_update_message(
        self, device_key: str, status: FirmwareUpdateStatus
    ) -> Message:
        """
        Make message from device firmware update status.

        :param device_key: Device key to which the firmware update status belongs to
        :type device_key: str
        :param status: Device firmware update status
        :type status: FirmwareUpdateStatus

        :returns: message
        :rtype: Message
        """
        topic = (
            self.FIRMWARE_UPDATE_STATUS_TOPIC_ROOT
            + self.DEVICE_PATH_PREFIX
            + device_key
        )
        payload = {"status": status.status.value}
        if status.error_code:
            payload["error"] = status.error_code.value

        message = Message(topic, json.dumps(payload))
        self.log.debug(f"Made {message} from {status} and {device_key}")
        return message

    def make_version_message(
        self, device_key: str, firmware_verison: str
    ) -> Message:
        """
        Make message from device firmware update version.

        :param device_key: Device key to which the firmware update belongs to
        :type device_key: str
        :param firmware_verison: Current firmware version
        :type firmware_verison: str

        :returns: message
        :rtype: Message
        """
        topic = (
            self.FIRMWARE_VERSION_UPDATE_TOPIC_ROOT
            + self.DEVICE_PATH_PREFIX
            + device_key
        )
        payload = str(firmware_verison)

        message = Message(topic, payload)
        self.log.debug(
            f"Made {message} from {firmware_verison} and {device_key}"
        )
        return message

    def is_firmware_install_command(self, message: Message) -> bool:
        """
        Check if received message is firmware install command.

        :param message: Message received
        :type message: Message

        :returns: is_firmware_install_command
        :rtype: bool
        """
        is_firmware_install_command = message.topic.startswith(
            self.FIRMWARE_UPDATE_INSTALL_TOPIC_ROOT
        )
        self.log.debug(
            f"Is {message} firmware install command "
            f"message: {is_firmware_install_command}"
        )
        return is_firmware_install_command

    def is_firmware_abort_command(self, message: Message) -> bool:
        """
        Check if received message is firmware abort command.

        :param message: Message received
        :type message: Message

        :returns: is_firmware_abort_command
        :rtype: bool
        """
        is_firmware_abort_command = message.topic.startswith(
            self.FIRMWARE_UPDATE_ABORT_TOPIC_ROOT
        )
        self.log.debug(
            f"Is {message} firmware abort command "
            f"message: {is_firmware_abort_command}"
        )
        return is_firmware_abort_command

    def make_firmware_file_path(self, message: Message) -> str:
        """
        Extract file path from firmware install message.

        :param message: Message received
        :type message: Message

        :returns: firmware_file_path
        :rtype: str
        """
        payload = json.loads(message.payload)  # type: ignore
        firmware_file_path = payload["fileName"]
        self.log.debug(f"Made {firmware_file_path} from {message}")
        return firmware_file_path

    def extract_key_from_message(self, message: Message) -> str:
        """
        Return device key from message.

        :param message: Message received
        :type message: Message

        :returns: device_key
        :rtype: str
        """
        device_key = message.topic.split("/")[-1]
        self.log.debug(f"Made {device_key} from {message}")
        return device_key



"""Response for device registration request."""
#   Copyright 2019 WolkAbout Technology s.r.o.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
from dataclasses import dataclass
from dataclasses import field

from wolk_gateway_module.model.device_registration_response_result import (
    DeviceRegistrationResponseResult,
)


@dataclass
class DeviceRegistrationResponse:
    """
    Response for device registration request.

    Identified by device key and result,
    with an optional description of the error that occurred.

    :ivar key: Unique device key
    :vartype key: str
    :ivar result: Result of the registration process
    :vartype result: DeviceRegistrationResponseResult
    :ivar description: Description of error that occurred
    :vartype description: str
    """

    key: str
    result: DeviceRegistrationResponseResult
    description: str = field(default="")



"""Handling of outbound and inbound messages related to device data."""
#   Copyright 2019 WolkAbout Technology s.r.o.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
from abc import ABC
from abc import abstractmethod
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union

from wolk_gateway_module.model.actuator_command import ActuatorCommand
from wolk_gateway_module.model.actuator_status import ActuatorStatus
from wolk_gateway_module.model.alarm import Alarm
from wolk_gateway_module.model.configuration_command import (
    ConfigurationCommand,
)
from wolk_gateway_module.model.message import Message
from wolk_gateway_module.model.sensor_reading import SensorReading

Configuration = Dict[
    str,
    Union[
        int,
        float,
        bool,
        str,
        Tuple[int, int],
        Tuple[int, int, int],
        Tuple[float, float],
        Tuple[float, float, float],
        Tuple[str, str],
        Tuple[str, str, str],
    ],
]


class DataProtocol(ABC):
    """Parse inbound messages and serialize outbound messages."""

    @abstractmethod
    def get_inbound_topics_for_device(self, device_key: str) -> List[str]:
        """
        Return list of inbound topics for given device key.

        :param device_key: Device key for which to create topics
        :type device_key: str

        :returns: inbound_topics
        :rtype: list
        """
        pass

    @abstractmethod
    def is_actuator_get_message(self, message: Message) -> bool:
        """
        Check if message is actuator get command.

        :param message: Message received
        :type message: Message

        :returns: is_actuator_get_message
        :rtype: bool
        """
        pass

    @abstractmethod
    def is_actuator_set_message(self, message: Message) -> bool:
        """
        Check if message is actuator set command.

        :param message: Message received
        :type message: Message

        :returns: is_actuator_set_message
        :rtype: bool
        """
        pass

    @abstractmethod
    def is_configuration_get_message(self, message: Message) -> bool:
        """
        Check if message is configuration get command.

        :param message: Message received
        :type message: Message

        :returns: is_configuration_get_message
        :rtype: bool
        """
        pass

    @abstractmethod
    def is_configuration_set_message(self, message: Message) -> bool:
        """
        Check if message is configuration set command.

        :param message: Message received
        :type message: Message

        :returns: is_configuration_set_message
        :rtype: bool
        """
        pass

    @abstractmethod
    def extract_key_from_message(self, message: Message) -> str:
        """
        Extract device key from message.

        :param message: Message received
        :type message: Message

        :returns: device_key
        :rtype: str
        """
        pass

    @abstractmethod
    def make_actuator_command(self, message: Message) -> ActuatorCommand:
        """
        Make actuator command from message.

        :param message: Message received
        :type message: Message

        :returns: actuator_command
        :rtype: ActuatorCommand
        """
        pass

    @abstractmethod
    def make_configuration_command(
        self, message: Message
    ) -> ConfigurationCommand:
        """
        Make configuration command from message.

        :param message: Message received
        :type message: Message

        :returns: configuration_command
        :rtype: ConfigurationCommand
        """
        pass

    @abstractmethod
    def make_sensor_reading_message(
        self, device_key: str, sensor_reading: SensorReading
    ) -> Message:
        """
        Make message from sensor reading for device key.

        :param device_key: Device on which the sensor reading occurred
        :type device_key: str
        :param sensor_reading: Sensor reading data
        :type sensor_reading: SensorReading

        :returns: message
        :rtype: Message
        """
        pass

    @abstractmethod
    def make_sensor_readings_message(
        self,
        device_key: str,
        sensor_readings: List[SensorReading],
        timestamp: Optional[int] = None,
    ) -> Message:
        """
        Make message from multiple sensor readings for device key.

        :param device_key: Device on which the sensor reading occurred
        :type device_key: str
        :param sensor_readings: List of sensor readings data
        :type sensor_readings: List[SensorReading]
        :param timestamp: Timestamp
        :type timestamp: Optional[int]

        :returns: message
        :rtype: Message
        """
        pass

    @abstractmethod
    def make_alarm_message(self, device_key: str, alarm: Alarm) -> Message:
        """
        Make message from alarm for device key.

        :param device_key: Device on which the alarm occurred
        :type device_key: str
        :param alarm: Alarm data
        :type alarm: Alarm

        :returns: message
        :rtype: Message
        """
        pass

    @abstractmethod
    def make_actuator_status_message(
        self, device_key: str, actuator_status: ActuatorStatus
    ) -> Message:
        """
        Make message from actuator status for device key.

        :param device_key: Device on which the actuator status occurred
        :type device_key: str
        :param actuator_status: Actuator status data
        :type actuator_status: ActuatorStatus

        :returns: message
        :rtype: Message
        """
        pass

    @abstractmethod
    def make_configuration_message(
        self, device_key: str, configuration: Configuration
    ) -> Message:
        """
        Make message from configuration for device key.

        :param device_key: Device to which the configuration belongs to.
        :type device_key: str
        :param configuration: Current configuration data
        :type configuration: dict

        :returns: message
        :rtype: Message
        """
        pass



"""Available reading type names."""
#   Copyright 2019 WolkAbout Technology s.r.o.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
from enum import Enum
from enum import unique


@unique
class ReadingTypeName(Enum):
    """Enumeration of defined reading type names on WolkAbout IoT Platform."""

    GENERIC = "GENERIC"
    TEMPERATURE = "TEMPERATURE"
    PRESSURE = "PRESSURE"
    HUMIDITY = "HUMIDITY"
    BATTERY_VOLTAGE = "BATTERY_VOLTAGE"
    MOVEMENT = "MOVEMENT"
    LIGHT = "LIGHT"
    ACCELEROMETER = "ACCELEROMETER"
    GYROSCOPE = "GYROSCOPE"
    LOCATION = "LOCATION"
    HEART_RATE = "HEART_RATE"
    COUNT = "COUNT"
    BATTERY_POWER = "BATTERY_POWER"
    BREATHING_RATE = "BREATHING_RATE"
    CALORIES = "CALORIES"
    ELECTRIC_CURRENT = "ELECTRIC_CURRENT"
    POWER = "POWER"
    FLOOR_POSITION = "FLOOR_POSITION"
    FLUID_VOLUME = "FLUID_VOLUME"
    LENGHT = "LENGHT"
    MASS = "MASS"
    SOUND_LEVEL = "SOUND_LEVEL"
    SPEED = "SPEED"
    STRING = "STRING"
    SWITCH = "SWITCH"
    TIME = "TIME"
    MAGNETIC_FLUX_DENSITY = "MAGNETIC_FLUX_DENSITY"
    RADIATION = "RADIATION"
    FORCE = "FORCE"
    MEASURE = "MEASURE"
    ANGLE = "ANGLE"
    FREQUENCY = "FREQUENCY"
    MAGNETIC_FLUX = "MAGNETIC_FLUX"
    ELECTRIC_CAPACITY = "ELECTRIC_CAPACITY"
    ELECTRIC_RESISTANCE = "ELECTRIC_RESISTANCE"
    ELECTRIC_CHARGE = "ELECTRIC_CHARGE"
    ELECTRIC_MAGNETISM = "ELECTRIC_MAGNETISM"
    ELECTRIC_ENERGY = "ELECTRIC_ENERGY"
    ELECTRIC_INDUCTANCE = "ELECTRIC_INDUCTANCE"
    ELECTRIC_CONDUCTANCE = "ELECTRIC_CONDUCTANCE"
    LUMINOUS_FLUX = "LUMINOUS_FLUX"
    LUMINOUS_INTENSITY = "LUMINOUS_INTENSITY"
    ILLUMINANCE = "ILLUMINANCE"
    GENERIC_TEXT = "GENERIC_TEXT"
    GENERIC_BOOLEAN = "GENERIC_BOOLEAN"



"""Store data before publishing."""
#   Copyright 2019 WolkAbout Technology s.r.o.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
from abc import ABC
from abc import abstractmethod
from typing import List
from typing import Optional

from wolk_gateway_module.model.message import Message


class OutboundMessageQueue(ABC):
    """Responsible for storing messages before being sent to WolkGateway."""

    @abstractmethod
    def put(self, message: Message) -> bool:
        """
        Place a message in storage.

        :param message: Message to be stored
        :type message: Message

        :returns: result
        :rtype: bool
        """
        pass

    @abstractmethod
    def get(self) -> Optional[Message]:
        """
        Get the first message from storage without removing it.

        :returns: message
        :rtype: Message, None
        """
        pass

    @abstractmethod
    def remove(self, message: Message) -> bool:
        """
        Remove specific message from storage.

        :returns: result
        :rtype: bool
        """
        pass

    @abstractmethod
    def get_messages_for_device(self, device_key: str) -> List[Message]:
        """
        Return a list of messages that belong to a certain device.

        Does not remove from storage.

        :param device_key: Device identifier
        :type device_key: str

        :returns: messages
        :rtype: List[Message]
        """
        pass

    @abstractmethod
    def queue_size(self) -> int:
        """
        Return current number of messages in storage.

        :returns: size
        :rtype: int
        """
        pass



"""Stub method for setting new configuration values."""
#   Copyright 2019 WolkAbout Technology s.r.o.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
from typing import Dict
from typing import Tuple
from typing import Union


def handle_configuration(
    device_key: str,
    configuration: Dict[
        str,
        Union[
            int,
            float,
            bool,
            str,
            Tuple[int, int],
            Tuple[int, int, int],
            Tuple[float, float],
            Tuple[float, float, float],
            Tuple[str, str],
            Tuple[str, str, str],
        ],
    ],
) -> None:
    """
    Change device's configuration options.

    Must be implemented as non blocking.
    Must be implemented as thread safe.

    :param device_key: Device identifier
    :type device_key: str
    :param configuration: Configuration option reference:value pairs
    :type configuration: dict
    """
    pass



"""Tests for JsonStatusProtocol."""
#   Copyright 2019 WolkAbout Technology s.r.o.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
import json
import sys
import unittest

sys.path.append("..")  # noqa

from wolk_gateway_module.json_status_protocol import JsonStatusProtocol
from wolk_gateway_module.model.device_status import DeviceStatus
from wolk_gateway_module.model.message import Message


class JsonStatusProtocolTests(unittest.TestCase):
    """JsonStatusProtocol Tests."""

    DEVICE_PATH_PREFIX = "d/"
    DEVICE_STATUS_UPDATE_TOPIC_ROOT = "d2p/subdevice_status_update/"
    DEVICE_STATUS_RESPONSE_TOPIC_ROOT = "d2p/subdevice_status_response/"
    DEVICE_STATUS_REQUEST_TOPIC_ROOT = "p2d/subdevice_status_request/"
    LAST_WILL_TOPIC = "lastwill"

    def test_get_inbound_topics_for_device(self):
        """Test that returned list is correctfor given device key."""
        json_status_protocol = JsonStatusProtocol()
        device_key = "some_key"

        self.assertEqual(
            [
                self.DEVICE_STATUS_REQUEST_TOPIC_ROOT
                + self.DEVICE_PATH_PREFIX
                + device_key
            ],
            json_status_protocol.get_inbound_topics_for_device(device_key),
        )

    def test_is_device_status_request_message(self):
        """Test that message is device status request."""
        json_status_protocol = JsonStatusProtocol()
        message = Message(self.DEVICE_STATUS_REQUEST_TOPIC_ROOT)
        self.assertTrue(
            json_status_protocol.is_device_status_request_message(message)
        )

    def test_make_device_status_response_message(self):
        """Test that message is made correctly."""
        json_status_protocol = JsonStatusProtocol()
        device_key = "some_key"
        status = DeviceStatus.CONNECTED

        expected = Message(
            self.DEVICE_STATUS_RESPONSE_TOPIC_ROOT
            + self.DEVICE_PATH_PREFIX
            + device_key,
            json.dumps({"state": status.value}),
        )

        serialized = json_status_protocol.make_device_status_response_message(
            status, device_key
        )

        self.assertEqual(expected, serialized)

    def test_make_device_status_update_message(self):
        """Test that message is made correctly."""
        json_status_protocol = JsonStatusProtocol()
        device_key = "some_key"
        status = DeviceStatus.OFFLINE

        expected = Message(
            self.DEVICE_STATUS_UPDATE_TOPIC_ROOT
            + self.DEVICE_PATH_PREFIX
            + device_key,
            json.dumps({"state": status.value}),
        )

        serialized = json_status_protocol.make_device_status_update_message(
            status, device_key
        )

        self.assertEqual(expected, serialized)

    def test_make_last_will_message(self):
        """Test that message is made correctly from device keys."""
        json_status_protocol = JsonStatusProtocol()
        keys = ["device1", "device2", "device3"]

        expected = Message(self.LAST_WILL_TOPIC, json.dumps(keys))

        serialized = json_status_protocol.make_last_will_message(keys)

        self.assertEqual(expected, serialized)


if __name__ == "__main__":
    unittest.main()



"""Configuration template for registering device on WolkAbout IoT Platform."""
#   Copyright 2019 WolkAbout Technology s.r.o.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
from typing import Dict
from typing import List
from typing import Optional
from typing import Union

from wolk_gateway_module.model.data_type import DataType


class ConfigurationTemplate:
    """
    Configuration template for registering device on Platform.

    :ivar data_type: Configuration data type
    :vartype data_type: DataType
    :ivar default_value: Default value of configuration
    :vartype default_value: str or None
    :ivar description: Description of configuration
    :vartype description: str or None
    :ivar labels: Labels of fields when data size > 1
    :vartype labels: List[str] or None
    :ivar maximum: Maximum configuration value
    :vartype maximum: int
    :ivar minimum: Minimum configuration value
    :vartype minimum: int
    :ivar name: Configuration name
    :vartype name: str
    :ivar reference: Unique configuration reference
    :vartype reference: str
    :ivar size: Data size
    :vartype size: int
    """

    def __init__(
        self,
        name: str,
        reference: str,
        data_type: DataType,
        description: Optional[str] = None,
        size: int = 1,
        labels: Optional[List[str]] = None,
        default_value: Optional[str] = None,
        minimum: Optional[Union[int, float]] = None,
        maximum: Optional[Union[int, float]] = None,
    ):
        """
        Configuration template for device registration request.

        :param name: Configuration name
        :type name: str
        :param reference: Configuration reference
        :type reference: str
        :param data_type: Configuration data type
        :type data_type: DataType
        :param description: Configuration description
        :type description: Optional[str]
        :param size: Configuration data size (max 3)
        :type size: Optional[int]
        :param labels: List of string lables when data size > 1
        :type labels: Optional[List[str]]
        :param default_value: Default configuration value
        :type default_value: Optional[str]
        :param minimum: Minimum configuration value
        :type minimum: Optional[Union[int, float]]
        :param maximum: Maximum configuration value
        :type maximum: Optional[Union[int, float]]
        """
        self.name: str = name
        self.reference: str = reference
        self.description: Optional[str] = description
        self.minimum: Optional[Union[int, float]] = minimum
        self.maximum: Optional[Union[int, float]] = maximum
        self.default_value: Optional[str] = default_value
        if size < 1 or size > 3:
            raise ValueError("Size can only be 1, 2 or 3")
        if size == 1:
            self.size: int = 1
            self.labels: Optional[List[str]] = None
        else:
            self.size = size
            if not labels:
                raise ValueError("Lables must be provided for size > 1")
            self.labels = labels
        if not isinstance(data_type, DataType):
            raise ValueError("Invalid data type given")
        self.data_type: DataType = data_type

    def __repr__(self) -> str:
        """
        Make string representation of configuration template.

        :returns: representation
        :rtype: str
        """
        return (
            f"ConfigurationTemplate(name='{self.name}', "
            f"reference='{self.reference}', description='{self.description}', "
            f"data_type='{self.data_type}', "
            f"minimum='{self.minimum}', maximum='{self.maximum}')"
            f"default_value='{self.default_value}', "
            f"size='{self.size}', "
            f"labels='{self.labels}')"
        )

    def to_dto(self) -> Dict[str, Union[int, str, float, List[str]]]:
        """
        Create data transfer object used for registration.

        :returns: dto
        :rtype: Dict[str, Union[str, int, float, List[str]]]
        """
        dto: Dict[str, Union[str, int, float, List[str]]] = {
            "name": self.name,
            "reference": self.reference,
            "dataType": self.data_type.name,
        }

        if self.size != 1 and self.labels is not None:
            dto.update({"size": self.size})
            dto["labels"] = self.labels
        else:
            dto["labels"] = []

        if self.minimum is not None and self.maximum is not None:
            dto["minimum"] = self.minimum
            dto["maximum"] = self.maximum

        if self.default_value:
            dto["defaultValue"] = str(self.default_value)

        return dto



"""Actuator status model."""
#   Copyright 2019 WolkAbout Technology s.r.o.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
from dataclasses import dataclass
from dataclasses import field
from typing import Optional
from typing import Union

from wolk_gateway_module.model.actuator_state import ActuatorState


@dataclass
class ActuatorStatus:
    """
    Holds information of a devices actuator current status.

    :ivar reference: Device actuator's reference as defined in device template
    :vartype reference: str
    :ivar state: Actuator's current state
    :vartype state: ActuatorState
    :ivar value: Current value of actuator, None only for error state
    :vartype value: Optional[Union[bool, int, float, str]]
    """

    reference: str
    state: ActuatorState
    value: Optional[Union[bool, int, float, str]] = field(default=None)



"""Device template used for registering device on WolkAbout IoT Platform."""
#   Copyright 2019 WolkAbout Technology s.r.o.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
from dataclasses import dataclass
from dataclasses import field
from typing import Dict
from typing import List

from wolk_gateway_module.model.actuator_template import ActuatorTemplate
from wolk_gateway_module.model.alarm_template import AlarmTemplate
from wolk_gateway_module.model.configuration_template import (
    ConfigurationTemplate,
)
from wolk_gateway_module.model.sensor_template import SensorTemplate


@dataclass
class DeviceTemplate:
    """
    Contains information required for registering device on Platform.

    A device template consists of lists of templates (actuator, alarm, sensor, configuration)
    that represent what data the device is expected to send and receive.
    All references of a device must be unique.

    Other than data feed templates, there is a ``supports_firmware_update`` parameter
    that specifies if this device has the capability to perform firmware updates.

    Finally, there are type, connectivity and firmware update parameters that are dictionaries
    that will contain more attributes to group together devices, but are unused at this moment.

    :ivar actuators: List of actuators on device
    :vartype actuators: List[ActuatorTemplate]
    :ivar alarms: List of alarms on device
    :vartype alarms: List[AlarmTemplate]
    :ivar configurations: List of configurations on device
    :vartype configurations: List[ConfigurationTemplate]
    :ivar connectivity_parameters: Device's connectivity parameters
    :vartype connectivity_parameters: Dict[str, Union[str, int, float, bool]]
    :ivar supports_firmware_update: Is firmware update enabled for this device
    :vartype supports_firmware_update: bool
    :ivar sensors: List of sensors on device
    :vartype sensors: List[SensorTemplate]
    :ivar type_parameters: Device's type parameters
    :vartype type_parameters: Dict[str, Union[str, int, float, bool]]
    :ivar firmware_update_parameters: Device's firmware update parameters
    :vartype firmware_update_parameters: Dict[str, Union[str, int, float, bool]]
    """

    actuators: List[ActuatorTemplate] = field(default_factory=list)
    alarms: List[AlarmTemplate] = field(default_factory=list)
    configurations: List[ConfigurationTemplate] = field(default_factory=list)
    sensors: List[SensorTemplate] = field(default_factory=list)
    supports_firmware_update: bool = field(default=False)
    type_parameters: Dict = field(default_factory=dict)
    connectivity_parameters: Dict = field(default_factory=dict)
    firmware_update_parameters: Dict = field(default_factory=dict)



"""Handling of inbound and outbound messages related to device status."""
#   Copyright 2019 WolkAbout Technology s.r.o.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
import json

from wolk_gateway_module.logger_factory import logger_factory
from wolk_gateway_module.model.device_status import DeviceStatus
from wolk_gateway_module.model.message import Message
from wolk_gateway_module.protocol.status_protocol import StatusProtocol


class JsonStatusProtocol(StatusProtocol):
    """Parse inbound messages and serialize device status messages."""

    DEVICE_PATH_PREFIX = "d/"
    DEVICE_STATUS_UPDATE_TOPIC_ROOT = "d2p/subdevice_status_update/"
    DEVICE_STATUS_RESPONSE_TOPIC_ROOT = "d2p/subdevice_status_response/"
    DEVICE_STATUS_REQUEST_TOPIC_ROOT = "p2d/subdevice_status_request/"
    LAST_WILL_TOPIC = "lastwill"

    def __init__(self) -> None:
        """Create object."""
        self.log = logger_factory.get_logger(str(self.__class__.__name__))

    def __repr__(self) -> str:
        """
        Make string representation of JsonStatusProtocol.

        :returns: representation
        :rtype: str
        """
        return "JsonStatusProtocol()"

    def get_inbound_topics_for_device(self, device_key: str) -> list:
        """
        Return list of inbound topics for given device key.

        :param device_key: Device key for which to create topics
        :type device_key: str

        :returns: inbound_topics
        :rtype: list
        """
        inbound_topics = [
            self.DEVICE_STATUS_REQUEST_TOPIC_ROOT
            + self.DEVICE_PATH_PREFIX
            + device_key
        ]
        self.log.debug(f"Inbound topics for {device_key} : {inbound_topics}")

        return inbound_topics

    def is_device_status_request_message(self, message: Message) -> bool:
        """
        Check if message is device status request.

        :param message: Message received
        :type message: Message

        :returns: is_device_status_request
        :rtype: bool
        """
        is_device_status_request = message.topic.startswith(
            self.DEVICE_STATUS_REQUEST_TOPIC_ROOT
        )
        self.log.debug(
            f"Is {message} device status request "
            f"message: {is_device_status_request}"
        )

        return is_device_status_request

    def make_device_status_response_message(
        self, device_status: DeviceStatus, device_key: str
    ) -> Message:
        """
        Make message from device status response.

        :param device_key: Device to which the status belongs to
        :type device_key: str
        :param device_status: Device's current status
        :type device_status: DeviceStatus

        :returns: message
        :rtype: Message
        """
        message = Message(
            self.DEVICE_STATUS_RESPONSE_TOPIC_ROOT
            + self.DEVICE_PATH_PREFIX
            + device_key,
            json.dumps({"state": device_status.value}),
        )
        self.log.debug(f"Made {message} from {device_status} and {device_key}")

        return message

    def make_device_status_update_message(
        self, device_status: DeviceStatus, device_key: str
    ) -> Message:
        """
        Make message from device status update.

        :param device_key: Device to which the status belongs to
        :type device_key: str
        :param device_status: Device's current status
        :type device_status: DeviceStatus

        :returns: message
        :rtype: Message
        """
        message = Message(
            self.DEVICE_STATUS_UPDATE_TOPIC_ROOT
            + self.DEVICE_PATH_PREFIX
            + device_key,
            json.dumps({"state": device_status.value}),
        )
        self.log.debug(f"Made {message} from {device_status} and {device_key}")

        return message

    def make_last_will_message(self, device_keys: list) -> Message:
        """
        Make last will message from list of device keys.

        :param device_keys: List of device keys
        :type device_keys: list(str)

        :returns: message
        :rtype: Message
        """
        message = Message(self.LAST_WILL_TOPIC, json.dumps(device_keys))
        self.log.debug(f"Made {message} from {device_keys}")

        return message

    def extract_key_from_message(self, message: Message) -> str:
        """
        Extract device key from message.

        :param message: Message received
        :type message: Message

        :returns: device_key
        :rtype: str
        """
        device_key = message.topic.split("/")[-1]
        self.log.debug(f"Made {device_key} from {message}")

        return device_key



"""Module that provides connectivity for non-IP enabled devices to WolkAbout IoT Platform."""
#   Copyright 2020 WolkAbout Technology s.r.o.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
__version__ = "1.0.8"

from .connectivity.connectivity_service import ConnectivityService
from .interface.actuation_handler import handle_actuation
from .interface.actuator_status_provider import get_actuator_status
from .interface.configuration_handler import handle_configuration
from .interface.configuration_provider import get_configuration
from .interface.device_status_provider import get_device_status
from .interface.firmware_handler import FirmwareHandler
from .logger_factory import logging_config
from .model.actuator_command import ActuatorCommand
from .model.actuator_state import ActuatorState
from .model.actuator_status import ActuatorStatus
from .model.actuator_template import ActuatorTemplate
from .model.alarm import Alarm
from .model.alarm_template import AlarmTemplate
from .model.configuration_command import ConfigurationCommand
from .model.configuration_template import ConfigurationTemplate
from .model.data_type import DataType
from .model.device import Device
from .model.device_registration_request import DeviceRegistrationRequest
from .model.device_registration_response import DeviceRegistrationResponse
from .model.device_status import DeviceStatus
from .model.device_template import DeviceTemplate
from .model.firmware_update_status import FirmwareUpdateErrorCode
from .model.firmware_update_status import FirmwareUpdateState
from .model.firmware_update_status import FirmwareUpdateStatus
from .model.message import Message
from .model.reading_type import ReadingType
from .model.reading_type_measurement_unit import ReadingTypeMeasurementUnit
from .model.reading_type_name import ReadingTypeName
from .model.sensor_reading import SensorReading
from .model.sensor_template import SensorTemplate
from .persistence.outbound_message_queue import OutboundMessageQueue
from .protocol.data_protocol import DataProtocol
from .protocol.firmware_update_protocol import FirmwareUpdateProtocol
from .protocol.registration_protocol import RegistrationProtocol
from .protocol.status_protocol import StatusProtocol
from .wolk import Wolk

__all__ = [
    "Wolk",
    "logging_config",
    "handle_actuation",
    "get_actuator_status",
    "handle_configuration",
    "get_configuration",
    "get_device_status",
    "ActuatorState",
    "FirmwareHandler",
    "FirmwareUpdateState",
    "FirmwareUpdateStatus",
    "FirmwareUpdateErrorCode",
    "DataType",
    "ReadingType",
    "ReadingTypeMeasurementUnit",
    "ReadingTypeName",
    "SensorTemplate",
    "Device",
    "DeviceTemplate",
    "ActuatorTemplate",
    "AlarmTemplate",
    "ConfigurationTemplate",
    "ConnectivityService",
    "DataProtocol",
    "RegistrationProtocol",
    "StatusProtocol",
    "FirmwareUpdateProtocol",
    "OutboundMessageQueue",
    "Message",
    "ActuatorCommand",
    "ConfigurationCommand",
    "SensorReading",
    "Alarm",
    "ActuatorStatus",
    "DeviceRegistrationRequest",
    "DeviceRegistrationResponse",
    "DeviceStatus",
    "__version__",
]



"""Sensor templates used for registering device on WolkAbout IoT Platform."""
#   Copyright 2019 WolkAbout Technology s.r.o.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
from typing import Dict
from typing import Optional
from typing import Union

from wolk_gateway_module.model.data_type import DataType
from wolk_gateway_module.model.reading_type import ReadingType
from wolk_gateway_module.model.reading_type_measurement_unit import (
    ReadingTypeMeasurementUnit as Unit,
)
from wolk_gateway_module.model.reading_type_name import ReadingTypeName as Name


class SensorTemplate:
    """
    Sensor template for registering device on Platform.

    :ivar description: Description detailing this sensor
    :vartype description: str or None
    :ivar maximum: Maximum sensor value
    :vartype maximum: int or float or None
    :ivar minimum: Minimum sensor value
    :vartype minimum: int or float or None
    :ivar name: Name of sensor
    :vartype name: str
    :ivar reference: Unique sensor reference
    :vartype reference: str
    :ivar unit: Sensor reading type measurement name and unit
    :vartype unit: ReadingType
    """

    def __init__(
        self,
        name: str,
        reference: str,
        data_type: DataType = None,
        reading_type_name: Union[Name, str] = None,
        unit: Union[Unit, str] = None,
        description: Optional[str] = None,
        minimum: Optional[Union[int, float]] = None,
        maximum: Optional[Union[int, float]] = None,
    ):
        """
        Sensor template for device registration request.

        Define a reading type for sensors,
        either a generic type by specifying
        a ``DataType`` (numeric, boolean or string) or entering
        a predefined one by using the enumerations provided in
        ``ReadingTypeName`` and ``ReadingTypeMeasurementUnit``.

        Custom reading types that have been previously defined
        on WolkAbout IoT Platform can be used by passing string values
        for ``reading_type_name`` and ``unit``.

        :param name: Sensor name
        :type name: str
        :param reference: Sensor reference
        :type reference: str
        :param data_type: Sensor data type for generic reading type
        :type data_type: Optional[DataType]
        :param reading_type_name: Reading type name from defined enumeration or
         string for custom
        :type reading_type_name: Optional[Union[ReadingTypeName, str]]
        :param unit: Reading type measurement unit from defined enumeration
         or string for custom
        :type unit: Optional[Union[ReadingTypeMeasurementUnit, str]]
        :param description: Description detailing the sensor's specification
        :type description: Optional[str]
        :param minimum: Minimum sensor value
        :type minimum: Optional[Union[int, float]]
        :param maximum: Maximum sensor value
        :type maximum: Optional[Union[int, float]]
        """
        self.name = name
        self.reference = reference
        self.description = description
        self.minimum = minimum
        self.maximum = maximum

        if not (data_type or reading_type_name or unit):
            raise ValueError("Unable to create template")

        if data_type:
            if not isinstance(data_type, DataType):
                raise ValueError("Invalid data type given")
            if data_type == DataType.NUMERIC:
                self.unit = ReadingType(DataType.NUMERIC)
            elif data_type == DataType.BOOLEAN:
                self.unit = ReadingType(DataType.BOOLEAN)
            elif data_type == DataType.STRING:
                self.unit = ReadingType(DataType.STRING)
            return

        if not (
            reading_type_name
            and (isinstance(unit, str) or isinstance(unit, Unit))
        ):
            raise ValueError(
                "Both reading type name and unit must be provided"
            )
        self.unit = ReadingType(name=reading_type_name, unit=unit)

    def __repr__(self) -> str:
        """
        Make string representation of sensor template.

        :returns: representation
        :rtype: str
        """
        return (
            f"SensorTemplate(name='{self.name}', reference='{self.reference}',"
            f" description='{self.description}', unit='{self.unit}', "
            f"minimum='{self.minimum}', maximum='{self.maximum}')"
        )

    def to_dto(self) -> Dict[str, Union[str, int, float, Dict[str, str]]]:
        """
        Create data transfer object used for registration.

        :returns: dto
        :rtype: Dict[str, Union[str, int, float]]
        """
        dto: Dict[str, Union[str, int, float, Dict[str, str]]] = {
            "name": self.name,
            "reference": self.reference,
        }

        dto["description"] = self.description if self.description else ""

        dto["unit"] = (
            {
                "readingTypeName": self.unit.name.value,
                "symbol": self.unit.unit.value,
            }
            if (
                isinstance(self.unit.name, Name)
                and isinstance(self.unit.unit, Unit)
            )
            else {
                "readingTypeName": str(self.unit.name),
                "symbol": str(self.unit.unit),
            }
        )

        if self.minimum is not None and self.maximum is not None:
            dto["minimum"] = self.minimum
            dto["maximum"] = self.maximum

        return dto



"""Tests for JsonRegistrationProtocol."""
#   Copyright 2019 WolkAbout Technology s.r.o.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
import json
import sys
import unittest

sys.path.append("..")  # noqa

from wolk_gateway_module.json_registration_protocol import (
    JsonRegistrationProtocol,
)
from wolk_gateway_module.model.actuator_template import (
    ActuatorTemplate,
)
from wolk_gateway_module.model.alarm_template import AlarmTemplate
from wolk_gateway_module.model.configuration_template import (
    ConfigurationTemplate,
)
from wolk_gateway_module.model.data_type import DataType
from wolk_gateway_module.model.device_registration_request import (
    DeviceRegistrationRequest,
)
from wolk_gateway_module.model.device_registration_response import (
    DeviceRegistrationResponse,
)
from wolk_gateway_module.model.device_registration_response_result import (
    DeviceRegistrationResponseResult,
)
from wolk_gateway_module.model.device_template import DeviceTemplate
from wolk_gateway_module.model.message import Message
from wolk_gateway_module.model.reading_type_measurement_unit import (
    ReadingTypeMeasurementUnit,
)
from wolk_gateway_module.model.reading_type_name import ReadingTypeName
from wolk_gateway_module.model.sensor_template import SensorTemplate


class JsonRegistrationProtocolTests(unittest.TestCase):
    """JsonRegistrationProtocol Tests."""

    DEVICE_PATH_PREFIX = "d/"
    CHANNEL_WILDCARD = "#"
    DEVICE_REGISTRATION_REQUEST_TOPIC_ROOT = "d2p/register_subdevice_request/"
    DEVICE_REGISTRATION_RESPONSE_TOPIC_ROOT = (
        "p2d/register_subdevice_response/"
    )

    def test_get_inbound_topics_for_device(self):
        """Test that returned list is correct for given device key."""
        json_registration_protocol = JsonRegistrationProtocol()
        device_key = "some_key"

        self.assertEqual(
            json_registration_protocol.get_inbound_topics_for_device(
                device_key
            ),
            [
                self.DEVICE_REGISTRATION_RESPONSE_TOPIC_ROOT
                + self.DEVICE_PATH_PREFIX
                + device_key
            ],
        )

    def test_extract_key_from_message(self):
        """Test that device key is extracted."""
        json_registration_protocol = JsonRegistrationProtocol()
        device_key = "some_device_key"

        message = Message(
            self.DEVICE_REGISTRATION_RESPONSE_TOPIC_ROOT
            + self.DEVICE_PATH_PREFIX
            + device_key
        )

        self.assertEqual(
            json_registration_protocol.extract_key_from_message(message),
            device_key,
        )

    def test_is_registration_response_message(self):
        """Test that message is device registration response."""
        json_registration_protocol = JsonRegistrationProtocol()

        message = Message(self.DEVICE_REGISTRATION_RESPONSE_TOPIC_ROOT)

        self.assertTrue(
            json_registration_protocol.is_registration_response_message(
                message
            )
        )

    def test_empty_device_registration_request(self):
        """Test registration request for empty device template."""
        json_registration_protocol = JsonRegistrationProtocol()

        device_template = DeviceTemplate()
        device_name = "device_name"
        device_key = "device_key"

        device_registration_request = DeviceRegistrationRequest(
            device_name, device_key, device_template
        )

        expected_payload = json.loads(
            "{"
            + '"name": "device_name",'
            + '"deviceKey": "device_key",'
            + '"defaultBinding": true,'
            + '"sensors": [],'
            + '"actuators": [],'
            + '"alarms": [],'
            + '"configurations": [],'
            + '"typeParameters": {},'
            + '"connectivityParameters": {},'
            + '"firmwareUpdateParameters": {"supportsFirmwareUpdate": false},'
            + '"firmwareUpdateType": ""'
            + "}"
        )

        message = json_registration_protocol.make_registration_message(
            device_registration_request
        )

        self.assertEqual(expected_payload, json.loads(message.payload))

    def test_simple_device_registration_request(self):
        """Test registration request for simple device template."""
        json_registration_protocol = JsonRegistrationProtocol()

        device_name = "simple_device"
        device_key = "simple_key"
        temperature_sensor = SensorTemplate(
            "Temperature",
            "T",
            reading_type_name=ReadingTypeName.TEMPERATURE,
            unit=ReadingTypeMeasurementUnit.CELSIUS,
            minimum=0,
            maximum=100,
            description="A temperature sensor",
        )

        sensors = [temperature_sensor]

        device_template = DeviceTemplate(sensors=sensors)

        device_registration_request = DeviceRegistrationRequest(
            device_name, device_key, device_template
        )

        expected_payload = json.loads(
            "{"
            + '"name": "simple_device",'
            + '"deviceKey": "simple_key",'
            + '"defaultBinding": true,'
            + '"sensors": ['
            + '{"name": "Temperature", '
            + '"reference": "T", '
            + '"unit": {"readingTypeName": "TEMPERATURE", "symbol": "℃"}, '
            + '"description": "A temperature sensor", '
            + '"minimum": 0, '
            + '"maximum": 100}],'
            + '"actuators": [],'
            + '"alarms": [],'
            + '"configurations": [],'
            + '"typeParameters": {},'
            + '"connectivityParameters": {},'
            + '"firmwareUpdateParameters": {"supportsFirmwareUpdate": false},'
            + '"firmwareUpdateType": ""'
            + "}"
        )

        message = json_registration_protocol.make_registration_message(
            device_registration_request
        )

        self.assertEqual(expected_payload, json.loads(message.payload))

    def test_full_device_registration_request(self):
        """Test registration request for full device template."""
        json_registration_protocol = JsonRegistrationProtocol()

        device_name = "full_device"
        device_key = "full_key"
        temperature_sensor = SensorTemplate(
            "Temperature",
            "T",
            reading_type_name=ReadingTypeName.TEMPERATURE,
            unit=ReadingTypeMeasurementUnit.CELSIUS,
            minimum=0,
            maximum=100,
            description="A temperature sensor",
        )
        pressure_sensor = SensorTemplate(
            "Pressure",
            "P",
            reading_type_name=ReadingTypeName.PRESSURE,
            unit=ReadingTypeMeasurementUnit.MILLIBAR,
            description="A pressure sensor",
            minimum=300,
            maximum=1200,
        )
        humidity_sensor = SensorTemplate(
            "Humidity",
            "H",
            reading_type_name=ReadingTypeName.HUMIDITY,
            unit=ReadingTypeMeasurementUnit.PERCENT,
            description="A humidity sensor",
            minimum=0,
            maximum=100,
        )

        accelerometer_sensor = SensorTemplate(
            "Accelerometer",
            "ACL",
            reading_type_name=ReadingTypeName.ACCELEROMETER,
            unit=ReadingTypeMeasurementUnit.METRES_PER_SQUARE_SECOND,
            description="An accelerometer sensor",
            minimum=0,
            maximum=100,
        )

        sensors = [
            temperature_sensor,
            pressure_sensor,
            humidity_sensor,
            accelerometer_sensor,
        ]

        high_humidity_alarm = AlarmTemplate(
            "High Humidity", "HH", "High humidity has been detected"
        )

        alarms = [high_humidity_alarm]

        slider_actuator = ActuatorTemplate(
            "Slider", "SL", data_type=DataType.NUMERIC, minimum=0, maximum=100
        )

        switch_actuator = ActuatorTemplate(
            "Switch", "SW", data_type=DataType.BOOLEAN
        )

        actuators = [switch_actuator, slider_actuator]

        configuration_1 = ConfigurationTemplate(
            "configuration_1",
            "config_1",
            DataType.NUMERIC,
            minimum=0,
            maximum=100,
        )
        configuration_2 = ConfigurationTemplate(
            "configuration_2", "config_2", DataType.BOOLEAN
        )
        configuration_3 = ConfigurationTemplate(
            "configuration_3", "config_3", DataType.STRING
        )
        configuration_4 = ConfigurationTemplate(
            "configuration_4",
            "config_4",
            DataType.STRING,
            size=3,
            labels="a,b,c",
        )

        configurations = [
            configuration_1,
            configuration_2,
            configuration_3,
            configuration_4,
        ]

        device_template = DeviceTemplate(
            actuators, alarms, configurations, sensors, True
        )

        device_registration_request = DeviceRegistrationRequest(
            device_name, device_key, device_template
        )

        expected_payload = json.loads(
            "{"
            + '"name": "full_device",'
            + '"deviceKey": "full_key",'
            + '"defaultBinding": true,'
            + '"typeParameters": {},'
            + '"connectivityParameters": {},'
            + '"firmwareUpdateType": "DFU",'
            + '"sensors": ['
            + "  {"
            + '    "name": "Temperature",'
            + '    "reference": "T",'
            + '    "description": "A temperature sensor",'
            + '    "unit": {'
            + '      "readingTypeName": "TEMPERATURE",'
            + '      "symbol": "℃"'
            + "    },"
            + '    "minimum": 0,'
            + '    "maximum": 100'
            + "  },"
            + "  {"
            + '    "name": "Pressure",'
            + '    "reference": "P",'
            + '    "description": "A pressure sensor",'
            + '    "unit": {'
            + '      "readingTypeName": "PRESSURE",'
            + '      "symbol": "mb"'
            + "    },"
            + '    "minimum": 300,'
            + '    "maximum": 1200'
            + "  },"
            + "  {"
            + '    "name": "Humidity",'
            + '    "reference": "H",'
            + '    "description": "A humidity sensor",'
            + '    "unit": {'
            + '      "readingTypeName": "HUMIDITY",'
            + '      "symbol": "%"'
            + "    },"
            + '    "minimum": 0,'
            + '    "maximum": 100'
            + "  },"
            + "  {"
            + '    "name": "Accelerometer",'
            + '    "reference": "ACL",'
            + '    "description": "An accelerometer sensor",'
            + '    "unit": {'
            + '      "readingTypeName": "ACCELEROMETER",'
            + '      "symbol": "m/s²"'
            + "    },"
            + '    "minimum": 0,'
            + '    "maximum": 100'
            + "  }"
            + "],"
            + '"actuators": ['
            + "  {"
            + '    "name": "Switch",'
            + '    "reference": "SW",'
            + '    "unit": {'
            + '      "readingTypeName": "SWITCH(ACTUATOR)",'
            + '      "symbol": ""'
            + "    },"
            + '    "description": ""'
            + "  },"
            + "  {"
            + '    "name": "Slider",'
            + '    "reference": "SL",'
            + '    "unit": {'
            + '      "readingTypeName": "RANGE (ACTUATOR)",'
            + '      "symbol": "count"'
            + "    },"
            + '    "description": "",'
            + '    "minimum": 0,'
            + '    "maximum": 100'
            + "  }"
            + "],"
            + '"alarms": ['
            + "  {"
            + '    "name": "High Humidity",'
            + '    "reference": "HH",'
            + '    "description": "High humidity has been detected"'
            + "  }"
            + "],"
            + '"configurations": ['
            + "  {"
            + '    "name": "configuration_1",'
            + '    "reference": "config_1",'
            + '    "dataType": "NUMERIC",'
            + '    "labels": [],'
            + '    "minimum": 0,'
            + '    "maximum": 100'
            + "  },"
            + "  {"
            + '    "name": "configuration_2",'
            + '    "reference": "config_2",'
            + '    "dataType": "BOOLEAN",'
            + '    "labels": []'
            + "  },"
            + "  {"
            + '    "name": "configuration_3",'
            + '    "reference": "config_3",'
            + '    "dataType": "STRING",'
            + '    "labels": []'
            + "  },"
            + "  {"
            + '    "name": "configuration_4",'
            + '    "reference": "config_4",'
            + '    "dataType": "STRING",'
            + '    "size": 3,'
            + '    "labels": "a,b,c"'
            + "  }"
            + "],"
            + '"firmwareUpdateParameters": {'
            + '  "supportsFirmwareUpdate": true}'
            + "}"
        )

        message = json_registration_protocol.make_registration_message(
            device_registration_request
        )

        self.assertEqual(expected_payload, json.loads(message.payload))


def test_make_registration_response(self):
    """Test for valid response parsing."""
    json_registration_protocol = JsonRegistrationProtocol()
    message = Message(
        "", '{"payload":{"device_key":"some_key"}, "result":"OK"}'
    )

    expected = DeviceRegistrationResponse(
        "some_key", DeviceRegistrationResponseResult.OK
    )

    deserialized = json_registration_protocol.make_registration_response(
        message
    )

    self.assertEqual(expected, deserialized)


if __name__ == "__main__":
    unittest.main()



"""Alarm template used for registering device on WolkAbout IoT Platform."""
#   Copyright 2019 WolkAbout Technology s.r.o.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
from dataclasses import dataclass
from dataclasses import field
from typing import Dict
from typing import Optional


@dataclass
class AlarmTemplate:
    """
    Alarm template for registering device on WolkAbout IoT Platform.

    :ivar name: Alarm name
    :vartype name: str
    :ivar reference: Alarm reference
    :vartype reference: str
    :ivar description: Alarm description
    :vartype description: str
    """

    name: str
    reference: str
    description: Optional[str] = field(default="")

    def to_dto(self) -> Dict[str, str]:
        """
        Create data transfer object used for registration.

        :returns: dto
        :rtype: Dict[str, str]
        """
        return {
            "name": self.name,
            "reference": self.reference,
            "description": self.description if self.description else "",
        }



"""Everything needed for registering a device on WolkAbout IoT Platform."""
#   Copyright 2019 WolkAbout Technology s.r.o.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
from dataclasses import dataclass
from dataclasses import field
from typing import List

from wolk_gateway_module.model.device_template import DeviceTemplate


@dataclass
class Device:
    """
    Device identified by name and key, as well as its template.

    :ivar name: Device's name
    :vartype name: str
    :ivar key: Device's unique key
    :vartype key: str
    :ivar template: Device template that defines data the device will send and receive.
    :vartype template: DeviceTemplate
    """

    name: str
    key: str
    template: DeviceTemplate = field(default_factory=DeviceTemplate)

    def get_actuator_references(self) -> List[str]:
        """
        Get list of actuator references for device.

        :returns: actuator_references
        :rtype: List[str]
        """
        actuator_references = [
            actuator.reference for actuator in self.template.actuators
        ]
        return actuator_references

    def has_configurations(self) -> bool:
        """
        Return if device has configuration options.

        :returns: has_configurations
        :rtype: bool
        """
        return bool(self.template.configurations)

    def supports_firmware_update(self) -> bool:
        """
        Return if device supports firmware update.

        :returns: supports_firmware_update
        :rtype: bool
        """
        return self.template.supports_firmware_update



"""Tests for Wolk."""
#   Copyright 2020 WolkAbout Technology s.r.o.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
import logging
import sys
import unittest

sys.path.append("..")  # noqa

from wolk_gateway_module.connectivity.connectivity_service import (
    ConnectivityService,
)
from wolk_gateway_module.wolk import Wolk
from wolk_gateway_module.model.actuator_state import ActuatorState
from wolk_gateway_module.model.device import Device
from wolk_gateway_module.model.device_template import DeviceTemplate
from wolk_gateway_module.model.message import Message
from wolk_gateway_module.model.device_status import DeviceStatus
from wolk_gateway_module.interface.firmware_handler import FirmwareHandler
from wolk_gateway_module.model.firmware_update_status import (
    FirmwareUpdateStatus,
    FirmwareUpdateState,
)
from wolk_gateway_module.model.sensor_template import SensorTemplate
from wolk_gateway_module.model.data_type import DataType


class MockConnectivityService(ConnectivityService):

    _connected = False
    _topics = []

    def set_inbound_message_listener(self, on_inbound_message):
        pass

    def set_lastwill_message(self, message):
        self._lastwill = message

    def add_subscription_topics(self, topics):
        self._topics.extend(topics)

    def remove_topics_for_device(self, device_key):
        pass

    def connected(self):
        return self._connected

    def connect(self):
        return self._connected

    def reconnect(self):
        return self._connected

    def disconnect(self):
        pass

    def publish(self, message):
        return self._connected


actuator_1 = False
actuator_2 = 0
actuator_3 = "default"


def mock_actuator_handler(device_key, reference, value):
    if device_key == "key1":
        if reference == "REF1":
            global actuator_1
            actuator_1 = value
        elif reference == "REF2":
            global actuator_2
            actuator_2 = value
    elif device_key == "key2":
        if reference == "REF3":
            global actuator_3
            actuator_3 = value


def mock_actuator_status_provider(device_key, reference):
    if device_key == "key1":
        if reference == "REF1":
            global actuator_1
            return ActuatorState.READY, actuator_1
        elif reference == "REF2":
            global actuator_2
            return ActuatorState.READY, actuator_2
    elif device_key == "key2":
        if reference == "REF3":
            global actuator_3
            return ActuatorState.READY, actuator_3


configuration_1 = False
configuration_2 = 3.14
configuration_3 = ("string1", "string2")


def mock_configuration_provider(device_key):
    if device_key == "key1":
        global configuration_1
        global configuration_2
        return {
            "configuration_1": configuration_1,
            "configuration_2": configuration_2,
        }
    elif device_key == "key2":
        global configuration_3
        return {"configuration_3": configuration_3}


def mock_configuration_handler(device_key, configuration):
    if device_key == "key1":
        global configuration_1
        global configuration_2
        for key, value in configuration.items():
            if key == "configuration_1":
                configuration_1 = value
            elif key == "configuration_2":
                configuration_2 = value
    elif device_key == "key2":
        global configuration_3
        for key, value in configuration.items():
            if key == "configuration_3":
                configuration_3 = value


class MockFirmwareHandler(FirmwareHandler):
    def install_firmware(
        self, device_key: str, firmware_file_path: str
    ) -> None:
        """
        Handle the installation of the firmware file.

        Call `self.on_install_success(device_key)` to report success.
        Reporting success will also get new firmware version.

        If installation fails, call `self.on_install_fail(device_key, status)`
        where:
        `status = FirmwareUpdateStatus(
            FirmwareUpdateState.ERROR,
            FirmwareUpdateErrorCode.INSTALLATION_FAILED
        )`
        or use other values from `FirmwareUpdateErrorCode` if they fit better.

        :param device_key: Device for which the firmware command is intended
        :type device_key: str
        :param firmware_file_path: Path where the firmware file is located
        :type firmware_file_path: str
        """
        if device_key == "key1":
            self.on_install_success(device_key)

    def abort_installation(self, device_key: str) -> None:
        """
        Attempt to abort the firmware installation process for device.

        Call `self.on_install_fail(device_key, status)` to report if
        the installation process was able to be aborted with
        `status = FirmwareUpdateStatus(FirmwareUpdateState.ABORTED)`
        If unable to stop the installation process, no action is required.

        :param device_key: Device for which to abort installation
        :type device_key: str
        """
        if device_key == "key1":
            status = FirmwareUpdateStatus(FirmwareUpdateState.ABORTED)
            self.on_install_fail(device_key, status)

    def get_firmware_version(self, device_key: str) -> str:
        """
        Return device's current firmware version.

        :param device_key: Device identifier
        :type device_key: str
        :returns: version
        :rtype: str
        """
        if device_key == "key1":
            return "v1.0"
        elif device_key == "key2":
            return "v0.1"


class WolkTests(unittest.TestCase):
    """Wolk Tests."""

    def test_bad_status_provider(self):
        """Test that exception is raised for bad device status provider."""
        with self.assertRaises(RuntimeError):
            wolk = Wolk(  # noqa
                "host", 1883, "module_name", lambda a, b: a * b
            )

    def test_add_sensor_reading(self):
        """Test adding a sensor reading to storage and then publish."""
        wolk = Wolk(
            "host",
            1883,
            "module_name",
            lambda a: a,
            connectivity_service=MockConnectivityService(),
        )
        wolk.add_sensor_reading("device_key", "REF", 13)
        self.assertEqual(1, wolk.outbound_message_queue.queue_size())
        wolk.connectivity_service._connected = True
        wolk.publish()
        self.assertEqual(0, wolk.outbound_message_queue.queue_size())

    def test_add_alarm(self):
        """Test adding a alarm event to storage and then publish."""
        wolk = Wolk(
            "host",
            1883,
            "module_name",
            lambda a: a,
            connectivity_service=MockConnectivityService(),
        )
        wolk.add_alarm("device_key", "REF", False)
        self.assertEqual(1, wolk.outbound_message_queue.queue_size())
        wolk.connectivity_service._connected = True
        wolk.publish()
        self.assertEqual(0, wolk.outbound_message_queue.queue_size())

    def test_publish_actuator(self):
        """Test publishing actuator status."""
        wolk = Wolk(
            "host",
            1883,
            "module_name",
            lambda a: a,
            connectivity_service=MockConnectivityService(),
            actuation_handler=mock_actuator_handler,
            actuator_status_provider=mock_actuator_status_provider,
        )
        wolk.log.setLevel(logging.CRITICAL)
        wolk.publish_actuator_status("key1", "REF1")
        self.assertEqual(1, wolk.outbound_message_queue.queue_size())
        wolk.connectivity_service._connected = True
        wolk.publish()
        self.assertEqual(0, wolk.outbound_message_queue.queue_size())

    def test_publish_actuator_explicit(self):
        """Test publishing explicit actuator status."""
        wolk = Wolk(
            "host",
            1883,
            "module_name",
            lambda a: a,
            connectivity_service=MockConnectivityService(),
            actuation_handler=mock_actuator_handler,
            actuator_status_provider=mock_actuator_status_provider,
        )
        wolk.log.setLevel(logging.CRITICAL)
        wolk.publish_actuator_status("key1", "REF1", ActuatorState.READY, True)
        self.assertEqual(1, wolk.outbound_message_queue.queue_size())
        wolk.connectivity_service._connected = True
        wolk.publish()
        self.assertEqual(0, wolk.outbound_message_queue.queue_size())

    def test_receive_actuation(self):
        """Test receiving actuator set command."""
        wolk = Wolk(
            "host",
            1883,
            "module_name",
            lambda a: DeviceStatus.CONNECTED,
            connectivity_service=MockConnectivityService(),
            actuation_handler=mock_actuator_handler,
            actuator_status_provider=mock_actuator_status_provider,
        )
        wolk.log.setLevel(logging.CRITICAL)
        wolk.connectivity_service._connected = True
        message = Message("p2d/actuator_set/d/key1/r/REF2", '{"value": "3"}')
        wolk._on_inbound_message(message)

        global actuator_2
        self.assertEqual(3, actuator_2)

    def test_publish_configuration(self):
        """Test publishing configuration."""
        wolk = Wolk(
            "host",
            1883,
            "module_name",
            lambda a: a,
            connectivity_service=MockConnectivityService(),
            configuration_handler=mock_configuration_handler,
            configuration_provider=mock_configuration_provider,
        )
        wolk.log.setLevel(logging.CRITICAL)
        wolk.publish_configuration("key1")
        self.assertEqual(1, wolk.outbound_message_queue.queue_size())
        wolk.connectivity_service._connected = True
        wolk.publish()
        self.assertEqual(0, wolk.outbound_message_queue.queue_size())

    def test_receive_configuration(self):
        """Test receiving configuration set command."""
        wolk = Wolk(
            "host",
            1883,
            "module_name",
            lambda a: DeviceStatus.CONNECTED,
            connectivity_service=MockConnectivityService(),
            configuration_handler=mock_configuration_handler,
            configuration_provider=mock_configuration_provider,
        )
        wolk.log.setLevel(logging.CRITICAL)
        wolk.connectivity_service._connected = True
        message = Message(
            "p2d/configuration_set/d/key1",
            '{"configuration_1": "true", "configuration_2": "4.2"}',
        )
        wolk._on_inbound_message(message)

        self.assertEqual(True, configuration_1)
        self.assertEqual(4.2, configuration_2)

        message = Message(
            "p2d/configuration_set/d/key2",
            '{"configuration_3": "newstring1,newstring2"}',
        )
        wolk._on_inbound_message(message)

        self.assertEqual(("newstring1", "newstring2"), configuration_3)

    def test_publish_device_status(self):
        """Test sending device status."""
        wolk = Wolk(
            "host",
            1883,
            "module_name",
            lambda device_key: DeviceStatus.CONNECTED,
            connectivity_service=MockConnectivityService(),
        )
        wolk.log.setLevel(logging.CRITICAL)
        wolk.publish_device_status("key1")
        self.assertEqual(1, wolk.outbound_message_queue.queue_size())
        wolk.connectivity_service._connected = True
        wolk.publish()
        self.assertEqual(0, wolk.outbound_message_queue.queue_size())

    def test_publish_device_status_explicit(self):
        """Test sending explicit device status."""
        wolk = Wolk(
            "host",
            1883,
            "module_name",
            lambda device_key: DeviceStatus.CONNECTED,
            connectivity_service=MockConnectivityService(),
        )
        wolk.log.setLevel(logging.CRITICAL)
        wolk.publish_device_status("key1", DeviceStatus.OFFLINE)
        self.assertEqual(1, wolk.outbound_message_queue.queue_size())
        wolk.connectivity_service._connected = True
        wolk.publish()
        self.assertEqual(0, wolk.outbound_message_queue.queue_size())

    def test_on_status_request(self):
        """Test receiving device status request."""
        wolk = Wolk(
            "host",
            1883,
            "module_name",
            lambda device_key: DeviceStatus.CONNECTED,
            connectivity_service=MockConnectivityService(),
        )
        wolk.log.setLevel(logging.CRITICAL)
        message = Message("p2d/subdevice_status_request/d/key1")
        wolk._on_inbound_message(message)

        self.assertEqual(1, wolk.outbound_message_queue.queue_size())

    def test_firmware_update(self):
        """Test receiving firmware installation command."""
        wolk = Wolk(
            "host",
            1883,
            "module_name",
            lambda device_key: DeviceStatus.CONNECTED,
            connectivity_service=MockConnectivityService(),
            firmware_handler=MockFirmwareHandler(),
        )
        wolk.log.setLevel(logging.CRITICAL)

        message = Message(
            "p2d/firmware_update_install/d/key1", '{ "fileName": "some_path"}'
        )

        wolk._on_inbound_message(message)
        self.assertEqual(3, wolk.outbound_message_queue.queue_size())

    def test_firmware_abort(self):
        """Test receiving firmware abort command."""
        wolk = Wolk(
            "host",
            1883,
            "module_name",
            lambda device_key: DeviceStatus.CONNECTED,
            connectivity_service=MockConnectivityService(),
            firmware_handler=MockFirmwareHandler(),
        )
        wolk.log.setLevel(logging.CRITICAL)

        message = Message("p2d/firmware_update_abort/d/key1")

        wolk._on_inbound_message(message)
        self.assertEqual(1, wolk.outbound_message_queue.queue_size())

    def test_add_simple_device(self):
        """Test adding a simple device."""
        wolk = Wolk(
            "host",
            1883,
            "module_name",
            lambda device_key: DeviceStatus.CONNECTED,
            connectivity_service=MockConnectivityService(),
        )
        wolk.log.setLevel(logging.CRITICAL)

        sensor1 = SensorTemplate("sensor1", "s1", DataType.NUMERIC)

        device_template = DeviceTemplate(sensors=[sensor1])

        device = Device("device1", "key1", device_template)

        wolk.add_device(device)

        self.assertEqual(1, wolk.outbound_message_queue.queue_size())


if __name__ == "__main__":
    unittest.main()
